import datetime
import os
import sys
sys.path.append(os.path.abspath(os.path.dirname(__file__) + '/' + '../..'))

import numpy as np
import pandas as pd
import seaborn as sns
import scipy.stats
import statsmodels.api as sm
import statsmodels.formula.api as smf
import matplotlib.pyplot as plt

from python.tools import (
    clean_folder,
    get_mask_minimum_obs,
    get_OLS_results,
    get_PCA_results,
    summary_statistics_table,
    get_table_number_of_factors_GR,
    get_table_number_of_factors_EV
)

np.random.seed(20192019)

print datetime.datetime.now() 

# Total time: 12 hours
# max_iter = 200, eps = 1e-4,
# starting_vals = 50

################
## Parameters ##
################

file_name = './PGDP/input/individual_error_statistics_PGDP.csv'
output_folder = './PGDP/output'
var_names = ['SPFfor_Step{}'.format(ll) for ll in range(1, 5 + 1)]
min_T = 20
min_N = 20
Newey_West_lags = 4
max_consensus_lags = 3
max_factors = 4
eps = 1e-4
max_iter = 200
starting_vals = 50

clean_folder(output_folder)
clean_folder('{}/baseline'.format(output_folder))
clean_folder('{}/standardized'.format(output_folder))
clean_folder('{}/orthog_realiz'.format(output_folder))
clean_folder('{}/orthog_lag'.format(output_folder))
clean_folder('{}/orthog_realiz_lag'.format(output_folder))
clean_folder('{}/higher_min_N'.format(output_folder))
clean_folder('{}/outliers_removed'.format(output_folder))
clean_folder('{}/no_gaps'.format(output_folder))

#######################################
## Load data and construct variables ##
#######################################

df = pd.read_csv(file_name)

# Get datetime dates and sort values
df['datetime'] = df['DATE'].apply(lambda x: x.replace(':0', 'Q'))
df['datetime'] = pd.to_datetime(df['datetime'])

# Calculate forecast errors and lagged forecasts
for var in var_names:
    df['E_{}'.format(var)] = df['Realiz1'] - df[var]
    df['{}_LAG1'.format(var)] = df.groupby('ID').shift()[var]

# Calculate forecasts standardized
# by the cross-sectional standard deviation
df_temp = df.groupby('DATE').std()[var_names].reset_index()
for var in var_names:
    df_temp.rename(columns = {var: '{}_cs_str'.format(var)},
                   inplace = True)
df = pd.merge(df, df_temp, on = 'DATE', how = 'left')
for var in var_names:
    df['{}_std'.format(var)] = df[var] / df['{}_cs_str'.format(var)]

# Calculate consensus (average) forecasts
# and merge into the main dataframe
df_temp = df.groupby('DATE').mean().loc[:, var_names + ['{}_std'.format(name) for name in var_names]].reset_index()
for var in var_names + ['{}_std'.format(name) for name in var_names]:
    df_temp.rename(columns = {var: '{}_Consensus'.format(var)},
                        inplace = True)
df = pd.merge(df, df_temp, on = 'DATE', how = 'left')

# Calculate forecasts, adjusted for
# time fixed effects, by subtracting
# consensus forecast
for var in var_names:
    df['{}_cons_adj'.format(var)] = df[var] - df['{}_Consensus'.format(var)]

# Create subsample dataset with outliers removed
# by Tukey's rule
res_outliers = []
df_outliers_removed = df.copy()
for var in var_names:
    q_75 = df[var].quantile(0.75)
    q_25 = df[var].quantile(0.25)
    IQR = q_75 - q_25
    ub = q_75 + 1.5 * IQR
    lb = q_25 - 1.5 * IQR
    mask = (df[var] >= ub) | (df[var] <= lb)
    res_outliers.append({'var': var,
                         'frac_outliers': sum(mask) / float(df[var].count())})
    df_outliers_removed.loc[mask, var] = np.nan

mask = get_mask_minimum_obs(df_outliers_removed, 'SPFfor_Step1', min_N, min_T)
df_outliers_removed = df_outliers_removed.loc[mask, ]

res_outliers = pd.DataFrame(res_outliers)
res_outliers.to_csv('{}/outliers_removed/frac_outliers.csv'.format(output_folder), index = False)

# Create subsample dataset in which forecasters
# with gaps in survey participation of more than
# 5 years are removed

df_no_gaps = df.copy()
no_gaps = 0
no_observations_removed = 0
res_gaps = []

for ind in df['ID'].unique():
    mask = df['ID'] == ind
    df_ind = df.loc[mask, ].copy()
    df_ind = df_ind[['datetime', 'SPFfor_Step1']]
    df_ind = df_ind.dropna()
    df_ind['days_since_last'] = (df_ind['datetime'] - df_ind['datetime'].shift()).dt.days
    max_delta = df_ind['days_since_last'].max()
    res_gaps.append({'max_delta': max_delta,
                     'ID': ind})
    if max_delta > 365 * 5:
        no_gaps += 1
        no_observations_removed += df_ind.shape[0]
        df_no_gaps = df_no_gaps.loc[~mask, ]
res_gaps = pd.DataFrame(res_gaps)
res_gaps.to_csv('{}/gaps_forecasters.csv'.format(output_folder), index = False)
        
mask = get_mask_minimum_obs(df_no_gaps, 'SPFfor_Step1', min_N, min_T)
df_no_gaps = df_no_gaps.loc[mask, ]

# Create subsample with a higher required
# minimum number of survey participants
mask = get_mask_minimum_obs(df, 'SPFfor_Step1', min_N = 30, min_T = min_T)
df_high_min_N = df.loc[mask, ].copy()

######################
## Sample selection ##
######################

mask = get_mask_minimum_obs(df, 'SPFfor_Step1', min_N, min_T)

print '     Calculating summary statistics...'
stats_table = summary_statistics_table(df, mask, 
    var_names, '{}/summary_stats.txt'.format(output_folder))

df = df.loc[mask, ]

################################
## Decomposition of variation ##
################################

print '     Calculating variance decomposition...'

res = []
for var in var_names:
    indiv_FE = smf.ols('{} ~ C(ID)'.format(var), data = df).fit().rsquared
    time_FE = smf.ols('{} ~ C(DATE)'.format(var), data = df).fit().rsquared
    indiv_time_FE = smf.ols('{} ~ C(ID) + C(DATE)'.format(var), data = df).fit().rsquared
    temp = {'indiv_FE': indiv_FE,
            'time_FE': time_FE,
            'indiv_time_FE': indiv_time_FE,
            'variable': var}
    res.append(temp)
res = pd.DataFrame(res)
res.to_csv('{}/variance_decomposition.csv'.format(output_folder), index = False)

###############################################################
## OLS estimates of 1-factor model using consensus forecasts ##
###############################################################

print '     Calculating OLS results with consensus...'

get_OLS_results(df, var_names, realiz_name = 'Realiz1', 
    max_consensus_lags = max_consensus_lags, output_folder = '{}/baseline'.format(output_folder),
    Newey_West_lags = Newey_West_lags)

get_OLS_results(df, ['{}_std'.format(var) for var in var_names], realiz_name = 'Realiz1', 
    max_consensus_lags = max_consensus_lags, output_folder = '{}/standardized'.format(output_folder),
    Newey_West_lags = Newey_West_lags)

get_OLS_results(df, var_names, realiz_name = 'Realiz1', 
    max_consensus_lags = max_consensus_lags, output_folder = '{}/orthog_realiz'.format(output_folder),
    Newey_West_lags = Newey_West_lags,
    exog_regressors = ['Realiz1'])

get_OLS_results(df, var_names, realiz_name = 'Realiz1', 
    max_consensus_lags = max_consensus_lags, output_folder = '{}/orthog_lag'.format(output_folder),
    Newey_West_lags = Newey_West_lags,
    lagged_forecast = True)

get_OLS_results(df, var_names, realiz_name = 'Realiz1', 
    max_consensus_lags = max_consensus_lags, output_folder = '{}/orthog_realiz_lag'.format(output_folder),
    Newey_West_lags = Newey_West_lags,
    exog_regressors = ['Realiz1'],
    lagged_forecast = True)

##
## Subsample estimations
##

get_OLS_results(df_high_min_N, var_names, realiz_name = 'Realiz1', 
    max_consensus_lags = max_consensus_lags, output_folder = '{}/higher_min_N'.format(output_folder),
    Newey_West_lags = Newey_West_lags)

get_OLS_results(df_outliers_removed, var_names, realiz_name = 'Realiz1', 
    max_consensus_lags = max_consensus_lags, output_folder = '{}/outliers_removed'.format(output_folder),
    Newey_West_lags = Newey_West_lags)

get_OLS_results(df_no_gaps, var_names, realiz_name = 'Realiz1', 
    max_consensus_lags = max_consensus_lags, output_folder = '{}/no_gaps'.format(output_folder),
    Newey_West_lags = Newey_West_lags)

############################################
## PC estimates of multiple-factor models ##
############################################

print '     Calculating PCA results with multiple factors...'

get_PCA_results(df, var_names, realiz_name = 'Realiz1', 
    max_factors = max_factors + 2, output_folder = '{}/baseline'.format(output_folder), 
    Newey_West_lags = Newey_West_lags,
    eps = eps, max_iter = max_iter, starting_vals = starting_vals)

get_PCA_results(df, ['{}_std'.format(var) for var in var_names], realiz_name = 'Realiz1', 
    max_factors = max_factors + 2, output_folder = '{}/standardized'.format(output_folder), 
    Newey_West_lags = Newey_West_lags,
    eps = eps, max_iter = max_iter, starting_vals = starting_vals)

get_PCA_results(df, ['{}_cons_adj'.format(var) for var in var_names], realiz_name = 'Realiz1', 
    max_factors = max_factors + 2, output_folder = '{}/baseline'.format(output_folder), 
    Newey_West_lags = Newey_West_lags,
    eps = eps, max_iter = max_iter, starting_vals = starting_vals)

get_PCA_results(df, var_names, realiz_name = 'Realiz1', 
    max_factors = max_factors, output_folder = '{}/orthog_realiz'.format(output_folder), 
    Newey_West_lags = Newey_West_lags,
    eps = eps, max_iter = max_iter, starting_vals = starting_vals,
    exog_regressors = ['Realiz1'])

get_PCA_results(df, var_names, realiz_name = 'Realiz1', 
    max_factors = max_factors, output_folder = '{}/orthog_lag'.format(output_folder), 
    Newey_West_lags = Newey_West_lags,
    eps = eps, max_iter = max_iter, starting_vals = starting_vals,
    lagged_forecast = True)

get_PCA_results(df, var_names, realiz_name = 'Realiz1', 
    max_factors = max_factors, output_folder = '{}/orthog_realiz_lag'.format(output_folder), 
    Newey_West_lags = Newey_West_lags,
    eps = eps, max_iter = max_iter, starting_vals = starting_vals,
    exog_regressors = ['Realiz1'],
    lagged_forecast = True)

##
## Subsample estimations
##

get_PCA_results(df_high_min_N, var_names, realiz_name = 'Realiz1', 
    max_factors = max_factors + 2, output_folder = '{}/higher_min_N'.format(output_folder), 
    Newey_West_lags = Newey_West_lags,
    eps = eps, max_iter = max_iter, starting_vals = starting_vals)

get_PCA_results(df_high_min_N, ['{}_cons_adj'.format(var) for var in var_names], realiz_name = 'Realiz1', 
    max_factors = max_factors + 2, output_folder = '{}/higher_min_N'.format(output_folder), 
    Newey_West_lags = Newey_West_lags,
    eps = eps, max_iter = max_iter, starting_vals = starting_vals)

get_PCA_results(df_outliers_removed, var_names, realiz_name = 'Realiz1', 
    max_factors = max_factors, output_folder = '{}/outliers_removed'.format(output_folder), 
    Newey_West_lags = Newey_West_lags,
    eps = eps, max_iter = max_iter, starting_vals = starting_vals)

get_PCA_results(df_no_gaps, var_names, realiz_name = 'Realiz1', 
    max_factors = max_factors, output_folder = '{}/no_gaps'.format(output_folder), 
    Newey_West_lags = Newey_West_lags,
    eps = eps, max_iter = max_iter, starting_vals = starting_vals)

##
## Calculate optimal number of factors
##

get_table_number_of_factors_GR(output_folder = '{}/baseline'.format(output_folder), 
        output_str = 'baseline', number_factors = max_factors + 2)
get_table_number_of_factors_GR(output_folder = '{}/higher_min_N'.format(output_folder), 
        output_str = 'higher_min_N', number_factors = max_factors + 2)

get_table_number_of_factors_EV(output_folder = '{}/baseline'.format(output_folder), 
        output_str = 'baseline', number_factors = max_factors + 2)
get_table_number_of_factors_EV(output_folder = '{}/higher_min_N'.format(output_folder), 
        output_str = 'higher_min_N', number_factors = max_factors + 2)

print datetime.datetime.now()