import datetime
import os
import sys
sys.path.append(os.path.abspath(os.path.dirname(__file__) + '/' + '../..'))

import numpy as np
import pandas as pd
import statsmodels.api as sm
import statsmodels.formula.api as smf
import matplotlib.pyplot as plt
import re

from python.tools import (
    clean_folder,
    get_mask_minimum_obs
) 

################
## Parameters ##
################

input_folder = './graphs/input'
output_folder = './graphs/output/CG_regressions'
var_names = ['SPFfor_Step{}'.format(ll) for ll in range(1, 5 + 1)]
phi = 0.20
clean_folder(output_folder)

####################################
## Grab optimal number of factors ##
####################################

number_factors = pd.read_csv('{}/baseline/number_factors_GR_baseline.csv'.format(input_folder), 
                             index_col = 0)
number_factors = number_factors['Adjusted for Time F.E.'].to_dict()

##################
## Prepare data ##
##################

df = pd.read_csv('{}/individual_error_statistics_PGDP.csv'.format(input_folder))

# Calculate forecast errors and revisions
for ii in range(len(var_names)):
    var = var_names[ii]
    df['Efor_Step{}'.format(var[-1])] = df['Realiz1'] - df[var]
    if var != var_names[-1]:
        df['Rfor_Step{}'.format(var[-1])] = df[var] - df[var_names[ii + 1]]

# Get micro-level estimates
res = []
for ii in range(len(var_names)):
    if ii == len(var_names) - 1:
        continue
    var = var_names[ii]
    formula = 'Efor_Step{ii} ~ Rfor_Step{ii}'.format(ii = var[-1])
    df_temp = df.loc[:, ['Efor_Step{ii}'.format(ii = var[-1]), 
                         'Rfor_Step{ii}'.format(ii = var[-1]),
                         'ID']].copy()
    df_temp = df_temp.dropna()
    mod = smf.ols(formula, data = df_temp).fit()
    res.append({'coef': mod.params['Rfor_Step{}'.format(var[-1])],
                'var': var})
res_ind = pd.DataFrame(res)

# Get variance of forecast revisions
rev_names = ['Rfor_Step{}'.format(ii) for ii in range(1, 4 + 1)]
var_revisions = df.var()[rev_names]
res_ind['var_revs'] = var_revisions.values

# Get consensus forecasts
df_cons = df.groupby('DATE').mean().reset_index()
var_names = ['SPFfor_Step{}'.format(ii) for ii in range(1, 5 + 1)]
var_names.append('DATE')
df_cons = df_cons.loc[:, var_names]
df_realiz = df.loc[:, ['DATE', 'Realiz1']].dropna().groupby('DATE').first().reset_index()
df_cons = pd.merge(df_cons, df_realiz, on = 'DATE', how = 'left')

# Calculate forecast errors and revisions
var_names = ['SPFfor_Step{}'.format(ii) for ii in range(1, 5 + 1)]
for ii in range(len(var_names)):
    var = var_names[ii]
    df_cons['Efor_Step{}'.format(var[-1])] = df_cons['Realiz1'] - df_cons[var]
    if var != var_names[-1]:
        df_cons['Rfor_Step{}'.format(var[-1])] = df_cons[var] - df_cons[var_names[ii + 1]]

# Get consensus-level estimates
res = []
for ii in range(len(var_names)):
    if ii == len(var_names) - 1:
        continue
    var = var_names[ii]
    formula = 'Efor_Step{ii} ~ Rfor_Step{ii}'.format(ii = var[-1])
    df_temp = df_cons.loc[:, ['Efor_Step{ii}'.format(ii = var[-1]), 
                              'Rfor_Step{ii}'.format(ii = var[-1])]].copy()
    mod = smf.ols(formula, data = df_temp).fit()
    res.append({'coef': mod.params['Rfor_Step{}'.format(var[-1])],
                'var': var})
res_cons = pd.DataFrame(res)

#############################
## Calculate implied noise ##
#############################

res_ind['var_sigma_v'] = res_ind['var_revs'] * (- res_ind['coef'])
res_ind['var_sigma_v_corr'] = res_ind['var_revs'] * (- res_ind['coef']) / (1 - phi)
np.sqrt(res_ind['var_sigma_v'])

# Read in noise
for var in res_ind['var']:
    df_temp = pd.read_csv('{}/baseline/PCA_{}_number_factors_{}.csv'.format(input_folder, var, number_factors[var]))
    noise = df_temp['sigma_v'].mean()
    mask = res_ind['var'] == var
    res_ind.loc[mask, 'sigma_v'] = noise

# Save to LaTeX
res = pd.DataFrame()
res['Consensus'] = res_cons['coef']
res['Individual'] = res_ind['coef']
res['Revisions (S.D.)'] = res_ind['var_revs'] ** 0.5
res['Implied Measurement Error (S.D.), $\\phi = 0$'] = res_ind['var_sigma_v'] ** (0.5)
res['Implied Measurement Error (S.D.), $\\phi = 0.20$'] = res_ind['var_sigma_v_corr'] ** (0.5)
res['Estimated Noise (S.D.)'] = res_ind['sigma_v']
res.index = ['$h = {}$'.format(ii) for ii in range(0, 3 + 1)]
f1 = lambda x: '{:.2f}'.format(x)
latex = res.T.to_latex(column_format = 'lcccc', formatters = [f1, f1, f1, f1], escape = False)
with open('{}/CG_regressions.txt'.format(output_folder), 'w+') as text_file:
        text_file.write(latex)