import datetime
import os
import sys
sys.path.append(os.path.abspath(os.path.dirname(__file__) + '/' + '../..'))

import numpy as np
import pandas as pd
import statsmodels.api as sm
import statsmodels.formula.api as smf
import matplotlib.pyplot as plt
import re

from python.tools import (
    clean_folder,
    get_mask_minimum_obs
) 

################
## Parameters ##
################

input_folder = './graphs/input'
output_folder = './graphs/output/MSE_decomposition'
var_names_list = ['SPFfor_Step{}'.format(ii) for ii in range(1, 5 + 1)]
max_K = 4  # Number of lags in local projections when estimating
           # IRF of consensus forecast errors
max_L = 12 # Number of IRFs to be estimated

clean_folder(output_folder)

####################################
## Grab optimal number of factors ##
####################################

number_factors = pd.read_csv('{}/baseline/number_factors_GR_baseline.csv'.format(input_folder), 
                             index_col = 0)
number_factors = number_factors['Adjusted for Time F.E.'].to_dict()

################################
## Estimate irreducible error ##
################################

## Main loop for estimating the IRF
## of consensus forecast errors; the squared
## IRF gives the "composite bias coefficients"
## of Kucinskas and Peters (2022) that can be
## used to back variance of irreducible error

res = []

for var_name in var_names_list:

    # Get consensus dataset from the same
    # dataset that noise was estimated form
    df = pd.read_csv('{}/baseline/df_resid_PCA_{}_number_factors_{}.csv'.format(input_folder,
                                                                            var_name,
                                                                            number_factors[var_name]))
    df_cons = df.groupby('DATE').mean()

    # For h-step-ahead forecasts with h >= 2, we need
    # remove the first few IRFs because they are
    # mechanical; see Kucinskas and Peters (2022)
    offset = np.maximum(int(var_name[-1]) - 2, 0)
    
    # Estimate size of irreducible error
    # using the method of Kucinskas and Peters (2022)
    # with local projections

    # Contruct vatiables
    for ll in range(1, max_K + 1):
        df_cons['E_{}_LAG{}'.format(var_name, ll)] = df_cons['E_{}'.format(var_name)].shift(ll)

    for ll in range(offset, max_L + offset):
        df_cons['E_{}_LEAD{}'.format(var_name, ll)] = df_cons['E_{}'.format(var_name)].shift(-ll)

    # Estimate local projections
    temp = []

    # Main loop
    for ll in range(offset, max_L + offset):
        # Construct local-projections formula
        formula = 'E_{var_name}_LEAD{ll} ~ E_{var_name}_LAG1'.format(var_name = var_name,
                                                                ll = ll)
        for kk in range(2, max_K + 1):
            formula += ' + E_{}_LAG{}'.format(var_name, kk)
        
        # Estimate the model
        mod = smf.ols(formula, data = df_cons).fit()
        temp.append({'lag': ll + 1 - offset,
                    'IRF': mod.params['E_{}_LAG1'.format(var_name)]})

    temp = pd.DataFrame(temp)
    
    # Use the estimated IRFs to back out the irreducible error
    MSE = (df_cons['E_{}'.format(var_name)] ** 2).mean()
    IRF_sq_sum = (temp['IRF'] ** 2).sum()
    uncond_bias = (df_cons['E_{}'.format(var_name)].mean()) ** 2
    var_irreducible_error = (MSE - uncond_bias) / (1 + IRF_sq_sum) # Variance of irreducible error
    
    res.append({'var_name': var_name,
                'var_irreducible_error': var_irreducible_error,
                'uncond_bias': uncond_bias,
                'cond_bias' : IRF_sq_sum * var_irreducible_error,
                'MSE': MSE,
                'noise': 0.0,
                'level': 'consensus'})
res = pd.DataFrame(res)

############################################
## Get individual-level MSE decomposition ##
############################################

for var_name in var_names_list:
    # Read in dataset that was used in
    # the estimation of noise
    df = pd.read_csv('{}/baseline/df_resid_PCA_{}_number_factors_{}.csv'.format(input_folder,
                                                                            var_name,
                                                                            number_factors[var_name]))

    df['E_{}_sq'.format(var_name)] = df['E_{}'.format(var_name)] ** 2
    MSE = df.groupby('ID').mean()['E_{}_sq'.format(var_name)].mean()
    uncond_bias = (df.groupby('ID').mean()['E_{}'.format(var_name)] ** 2).mean()
    mask = (res['var_name'] == var_name) & (res['level'] == 'consensus')
    var_irreducible_error = res.loc[mask, 'var_irreducible_error'].values[0]
    
    # Read in noise
    df_noise = pd.read_csv('{}/baseline/PCA_{}_number_factors_{}.csv'.format(input_folder, 
                                                                             var_name, 
                                                                             number_factors[var_name]))
    
    # Average variance of noise across forecasters
    noise = (df_noise['sigma_v'] ** 2).mean()
    
    # Back out conditional bias
    cond_bias = MSE - var_irreducible_error - uncond_bias - noise
    
    # Save results
    res = res.append({'var_name': var_name,
                      'var_irreducible_error': var_irreducible_error,
                      'uncond_bias': uncond_bias,
                      'cond_bias' : cond_bias,
                      'MSE': MSE,
                      'noise': noise, 
                      'level': 'individual'}, 
                     ignore_index = True)

res.to_csv('{}/MSE_decomposition.csv'.format(output_folder), index = False)

###############
## Get table ##
###############

# Convert to percent
for var_name in ['uncond_bias', 'cond_bias', 'noise']:
    res[var_name] = 100 * (res[var_name] / res['MSE'])
res['irreducible_error'] = 100 * (res['var_irreducible_error'] / res['MSE'])

# Combine unconditional and conditional bias
res['bias'] = res['uncond_bias'] + res['cond_bias']

latex_template = """\\begin{tabular}{lccccccc}
\\toprule & \\multicolumn{4}{c}{Individual} & \\multicolumn{3}{c}{Consensus} \\\\
\\cmidrule(lr{0.5em}){2-5} \\cmidrule(lr{0.5em}){6-8}
& MSE & Bias & Noise & Irreducible & MSE & Bias & Irreducible \\\\
\cmidrule(lr{0.5em}){2-5} \cmidrule(lr{0.5em}){6-8}
"""
line_template = "$h = {h}$ & {mse_ind:.2f} & {bias_ind:.1f}\\% & {noise_ind:.1f}\\%"\
                "& {irred_ind:.1f}\\% & {mse_cons:.2f} & {bias_cons:.1f}\\%"\
                "& {irred_cons:.1f}\\% \\\\ \n"

for var_name in var_names_list:
    mask = res['var_name'] == var_name
    line = line_template.format(h = int(var_name[-1]) - 1,
                                mse_ind = res.loc[mask & (res['level'] == 'individual'), 'MSE'].values[0],
                                mse_cons = res.loc[mask & (res['level'] == 'consensus'), 'MSE'].values[0],
                                bias_ind = res.loc[mask & (res['level'] == 'individual'), 'bias'].values[0],
                                bias_cons = res.loc[mask & (res['level'] == 'consensus'), 'bias'].values[0],
                                noise_ind = res.loc[mask & (res['level'] == 'individual'), 'noise'].values[0],
                                irred_ind = res.loc[mask & (res['level'] == 'individual'), 'irreducible_error'].values[0],
                                irred_cons = res.loc[mask & (res['level'] == 'consensus'), 'irreducible_error'].values[0])
    latex_template = latex_template + line 
    if var_name == var_names_list[-1]:
        latex_template += '\\bottomrule \n\\end{tabular}'

with open('{}/MSE_decomposition.txt'.format(output_folder), 'w+') as text_file:
        text_file.write(latex_template)