import datetime
import os
import sys
sys.path.append(os.path.abspath(os.path.dirname(__file__) + '/' + '../..'))

import matplotlib as mpl

mpl.use('pgf')
pgf_with_latex = {                     
    "pgf.texsystem": "xelatex",        
    "pgf.rcfonts": False,
    "text.usetex": True,                
    "font.family": "Times New Roman",
    "pgf.preamble": [
        r"\usepackage{fontspec}",    
        r"\setmainfont{Times New Roman}",        
        r"\usepackage{unicode-math}",
        r"\setmathfont{XITS Math}"
        ]
    }    
mpl.rcParams.update(pgf_with_latex)

import numpy as np
import pandas as pd
import seaborn as sns
import scipy.stats
import statsmodels.api as sm
import statsmodels.formula.api as smf
import matplotlib.pyplot as plt

from python.tools import (
    clean_folder,
    get_mask_minimum_obs
) 

from graphs.tools import (
    collect_noise_results,
    get_table_noise_estimates,
    get_noise_cross_sectional_dispersion
)

################
## Parameters ##
################

input_folder = './graphs/input'
output_folder = './graphs/output/main'
dpi = 600
var_names = ['SPFfor_Step{}'.format(ii) for ii in range(1, 5 + 1)]
max_consensus_lags = 3
max_factors = 4
clean_folder(output_folder)

####################################
## Grab optimal number of factors ##
####################################

opt_number_factors = pd.read_csv('{}/baseline/number_factors_GR_baseline.csv'.format(input_folder), 
                             index_col = 0)
opt_number_factors = opt_number_factors['Adjusted for Time F.E.'].to_dict()

#########################################
## Get table of variance decomposition ##
#########################################

df = pd.read_csv('{}/variance_decomposition.csv'.format(input_folder))
df = df[['variable', 'indiv_FE', 'time_FE', 'indiv_time_FE']]
df.columns = ['Horizon', 'Forecaster', 'Time', 'Forecaster + Time']
for ii in df.index:
    df.iloc[ii, 0] = int(df.iloc[ii, 0][-1]) - 1
df.index = df['Horizon'].apply(lambda x: '$h = {}$'.format(x))
del df['Horizon']
f1 = lambda x: '{:.2f}'.format(x)
latex = df.to_latex(column_format = 'lccc', formatters = [f1, f1, f1], escape = False)

with open('{}/variance_decomposition.txt'.format(output_folder), 'w+') as text_file:
        text_file.write(latex)

########################################################
## Get table on number of observations (min_N, min_T) ##
########################################################

df = pd.read_csv('{}/individual_error_statistics_PGDP.csv'.format(input_folder))
min_T_list = range(0, 45, 10)
min_N_list = range(0, 45, 10)
res = []
for min_T in min_T_list:
    for min_N in min_N_list:
        mask = get_mask_minimum_obs(df, 'SPFfor_Step1', min_N, min_T)
        df_temp = df.loc[mask, ].copy()
        res.append({'min_T': min_T,
                    'min_N': min_N,
                    'total_NT': df_temp['SPFfor_Step1'].count(),
                    'total_N': df_temp['ID'].unique().shape[0],
                    'total_T': df_temp['DATE'].unique().shape[0]})
res = pd.DataFrame(res)

latex = pd.pivot_table(res, columns = 'min_T', index = 'min_N', values = 'total_NT')
latex.index.name = None
f1 = lambda x:'{:,}'.format(x)
latex = latex.to_latex(column_format = 'l' * 6, 
                       escape = False,
                       formatters = [f1, f1, f1, f1, f1])
latex = latex.replace('min_T', '\\diagbox{$N_\\text{min}$}{$T_\\text{min}$}')
with open('{}/min_N_T_observations.txt'.format(output_folder), 'w+') as text_file:
        text_file.write(latex)

##################################
## Get baseline noise estimates ##
##################################

df_OLS = collect_noise_results(var_names, '{}/baseline'.format(input_folder),
        model_type = 'OLS', number_lags = max_consensus_lags)
df_PCA = collect_noise_results(var_names, '{}/baseline'.format(input_folder),
        model_type = 'PCA', number_factors = max_factors)
get_table_noise_estimates(df_OLS, df_PCA, var_names, stat_name = 'noise', 
    output_folder = output_folder, file_name = 'baseline_noise')

#####################################
## Get estimates with higher min_N ##
#####################################

df_OLS = collect_noise_results(var_names, '{}/higher_min_N'.format(input_folder),
        model_type = 'OLS', number_lags = max_consensus_lags)
df_PCA = collect_noise_results(var_names, '{}/higher_min_N'.format(input_folder),
        model_type = 'PCA', number_factors = max_factors)
get_table_noise_estimates(df_OLS, df_PCA, var_names, stat_name = 'noise', 
    output_folder = output_folder, file_name = 'higher_min_N')

#############################################
## Get "no gaps" subsample noise estimates ##
#############################################

df_OLS = collect_noise_results(var_names, '{}/no_gaps'.format(input_folder),
        model_type = 'OLS', number_lags = max_consensus_lags)
df_PCA = collect_noise_results(var_names, '{}/no_gaps'.format(input_folder),
        model_type = 'PCA', number_factors = max_factors)
get_table_noise_estimates(df_OLS, df_PCA, var_names, stat_name = 'noise', 
    output_folder = output_folder, file_name = 'no_gaps_noise')

##############################################################
## Get contributions of noise to cross-sectional dispersion ##
##############################################################

df_OLS = get_noise_cross_sectional_dispersion(var_names, '{}/baseline'.format(input_folder),
        model_type = 'OLS', number_lags = max_consensus_lags)
df_PCA = get_noise_cross_sectional_dispersion(var_names, '{}/baseline'.format(input_folder),
        model_type = 'PCA', number_factors = max_factors)
get_table_noise_estimates(df_OLS, df_PCA, var_names, stat_name = 'ratio', 
    output_folder = output_folder, file_name = 'cross_sect_dispersion_noise',
    show_se = False)

######################################################
## Correlation between noise in different forecasts ##
######################################################                 

for var in var_names:
    df_temp = pd.read_csv('{}/baseline/PCA_{}_number_factors_{}.csv'.format(input_folder, var, opt_number_factors[var]))
    df_temp = df_temp[['ID', 'sigma_v']]
    df_temp.rename(columns = {'sigma_v': var},
                   inplace = True)
    if var == var_names[0]:
        df = df_temp
    else:
        df = pd.merge(df, df_temp, on = 'ID', how = 'left')

df = df.drop('ID', axis = 1).corr().round(2)
labels = ['$h = {}$'.format(hh) for hh in range(0, 4 + 1)]
df.index = labels
df.columns = labels
latex = df.to_latex(escape = False, column_format = 'lccccc',)
with open('{}/correlation_noise.txt'.format(output_folder), 'w+') as text_file:
        text_file.write(latex)

#########################################
## Graph of cross sectional dispersion ##
#########################################

df_temp = pd.read_csv('{}/baseline/df_resid_PCA_{}_number_factors_{}.csv'.format(input_folder, 'SPFfor_Step2', opt_number_factors['SPFfor_Step2']))
df_temp['DATE'] = pd.to_datetime(df_temp['DATE'].apply(lambda x: x.replace(':0', 'Q')))

df_full = pd.read_csv('{}/individual_error_statistics_PGDP.csv'.format(input_folder))
df_full['DATE'] = pd.to_datetime(df_full['DATE'].apply(lambda x: x.replace(':0', 'Q')))
df_dates = pd.Series(df_full['DATE'].unique()).sort_values()
df_dates = pd.DataFrame(df_dates)
df_dates.columns = ['DATE']

df_temp = df_temp.groupby('DATE').var()[['SPFfor_Step2_resid', 'SPFfor_Step2', 'Realiz1']].reset_index()
df_temp = pd.merge(df_dates, df_temp, on = 'DATE', how = 'left')

fig, ax1 = plt.subplots(figsize=(4.5, 3.0))
ax2 = ax1.twinx()
ax1.plot(df_temp['DATE'], df_temp['SPFfor_Step2'], '-b', alpha = 0.9, label = 'Disp.: Forecasts')
ax1.plot(df_temp['DATE'], df_temp['SPFfor_Step2_resid'] , '--r', alpha = 0.9, label = 'Disp.: Noise')
ax2.plot(df_temp['DATE'], 
         (df_temp['SPFfor_Step2_resid'] / df_temp['SPFfor_Step2']), ':k', alpha = 0.5)
ax1.legend(frameon = False, fontsize = 12, bbox_to_anchor=(1.039, -0.05),
           ncol=2)
ax1.set_ylabel('Variance', fontsize = 12)
ax2.set_ylabel('Noise / Forecasts', fontsize = 12)
plt.tight_layout()
plt.savefig("{}/dispersion.png".format(output_folder), dpi = dpi)
plt.savefig("{}/dispersion.pgf".format(output_folder), dpi = dpi)
plt.show()

###################################
## Get stringent noise estimates ##
###################################

df_OLS = collect_noise_results(var_names, '{}/orthog_realiz_lag'.format(input_folder),
        model_type = 'OLS', number_lags = max_consensus_lags)
df_PCA = collect_noise_results(var_names, '{}/orthog_realiz_lag'.format(input_folder),
        model_type = 'PCA', number_factors = max_factors)
get_table_noise_estimates(df_OLS, df_PCA, var_names, stat_name = 'noise', 
    output_folder = output_folder, file_name = 'stringent_noise')

df_OLS = collect_noise_results(var_names, '{}/orthog_realiz'.format(input_folder),
        model_type = 'OLS', number_lags = max_consensus_lags)
df_PCA = collect_noise_results(var_names, '{}/orthog_realiz'.format(input_folder),
        model_type = 'PCA', number_factors = max_factors)
get_table_noise_estimates(df_OLS, df_PCA, var_names, stat_name = 'noise', 
    output_folder = output_folder, file_name = 'stringent_realiz_only_noise')

df_OLS = collect_noise_results(var_names, '{}/orthog_lag'.format(input_folder),
        model_type = 'OLS', number_lags = max_consensus_lags)
df_PCA = collect_noise_results(var_names, '{}/orthog_lag'.format(input_folder),
        model_type = 'PCA', number_factors = max_factors)
get_table_noise_estimates(df_OLS, df_PCA, var_names, stat_name = 'noise', 
    output_folder = output_folder, file_name = 'stringent_lag_only_noise')

#####################################
## Get noise with outliers removed ##
#####################################

df_OLS = collect_noise_results(var_names, '{}/outliers_removed'.format(input_folder),
        model_type = 'OLS', number_lags = max_consensus_lags)
df_PCA = collect_noise_results(var_names, '{}/outliers_removed'.format(input_folder),
        model_type = 'PCA', number_factors = max_factors)
get_table_noise_estimates(df_OLS, df_PCA, var_names, stat_name = 'noise', 
    output_folder = output_folder, file_name = 'noise_outliers_removed')

#########################################
## Get noise on standardized forecasts ##
#########################################

horizons = {'SPFfor_Step1_std': 0,
            'SPFfor_Step2_std': 1,
            'SPFfor_Step3_std': 2,
            'SPFfor_Step4_std': 3,
            'SPFfor_Step5_std': 4}
df_OLS = collect_noise_results(['{}_std'.format(name) for name in var_names], '{}/standardized'.format(input_folder),
        model_type = 'OLS', number_lags = max_consensus_lags)
df_PCA = collect_noise_results(['{}_std'.format(name) for name in var_names], '{}/standardized'.format(input_folder),
        model_type = 'PCA', number_factors = max_factors)
get_table_noise_estimates(df_OLS, df_PCA, ['{}_std'.format(name) for name in var_names], stat_name = 'noise', 
    output_folder = output_folder, file_name = 'noise_standardized', horizons = horizons)

##############################################
## Get table for predictability regressions ##
##############################################

res = []
for var in var_names:
    df_temp = pd.read_csv('{}/baseline/predictability_PCA_{}_number_factors_{}.csv'.format(input_folder, var, opt_number_factors[var]))
    df_temp['t_stat'] = df_temp['gamma'] / df_temp['se_adj_NW']
    temp = {'\% $t \geq 1.96$': 100 * np.sum(df_temp['t_stat'] >= 1.96) / float(df_temp['t_stat'].count()),
            '\% $t \leq -1.96$': 100 * np.sum(df_temp['t_stat'] <= -1.96) / float(df_temp['t_stat'].count()),
            '\% $|t| \geq 1.96$': 100 * np.sum(np.abs(df_temp['t_stat']) >= 1.96) / float(df_temp['t_stat'].count()),
            'Med. $R^2$': df_temp['R2'].median(),
            'P90 $R^2$': df_temp['R2'].quantile(0.90),
            '$N$': df_temp['t_stat'].count(),
            'var': var}
    res.append(temp)
df_temp.head()
res = pd.DataFrame(res)
res.index = ['$h = {}$'.format(hh) for hh in range(0, 4 + 1)]
del res['var']
res = res[['$N$',
           '\% $t \geq 1.96$',
           '\% $t \leq -1.96$',
           '\% $|t| \geq 1.96$',
           'Med. $R^2$',
           'P90 $R^2$']]
f0 = lambda x: '{}'.format(x)
f1 = lambda x: '{:.2f}'.format(x)
latex = res.to_latex(column_format = 'lcccccc', formatters = [f0, f1, f1, f1, f1, f1], escape = False)
with open('{}/predictability_regs.txt'.format(output_folder), 'w+') as text_file:
        text_file.write(latex)

##################################
## Get autocorrelation of noise ##
##################################

df_OLS = collect_noise_results(var_names, '{}/baseline'.format(input_folder),
        model_type = 'OLS', number_lags = max_consensus_lags,
        within_variation = False)
df_PCA = collect_noise_results(var_names, '{}/baseline'.format(input_folder),
        model_type = 'PCA', number_factors = max_factors,
        within_variation = False)
get_table_noise_estimates(df_OLS, df_PCA, var_names, stat_name = 'autocorr_noise', 
    output_folder = output_folder, file_name = 'baseline_noise_autocorr',
    show_se = False)

#####################################
## Get standard deviation of noise ##
#####################################

# Baseline estimates
df_OLS = collect_noise_results(var_names, '{}/baseline'.format(input_folder),
        model_type = 'OLS', number_lags = max_consensus_lags,
        within_variation = False)
df_PCA = collect_noise_results(var_names, '{}/baseline'.format(input_folder),
        model_type = 'PCA', number_factors = max_factors,
        within_variation = False)
get_table_noise_estimates(df_OLS, df_PCA, var_names, stat_name = 'noise', 
    output_folder = output_folder, file_name = 'baseline_noise_std_dev',
    show_se = True)

# Higher min_N subsample
df_OLS = collect_noise_results(var_names, '{}/higher_min_N'.format(input_folder),
        model_type = 'OLS', number_lags = max_consensus_lags)
df_PCA = collect_noise_results(var_names, '{}/higher_min_N'.format(input_folder),
        model_type = 'PCA', number_factors = max_factors)
get_table_noise_estimates(df_OLS, df_PCA, var_names, stat_name = 'noise', 
    output_folder = output_folder, file_name = 'noise_higher_min_N_std_dev',
    show_se = True)

# Outliers removed estimates
df_OLS = collect_noise_results(var_names, '{}/outliers_removed'.format(input_folder),
        model_type = 'OLS', number_lags = max_consensus_lags,
        within_variation = False)
df_PCA = collect_noise_results(var_names, '{}/outliers_removed'.format(input_folder),
        model_type = 'PCA', number_factors = max_factors,
        within_variation = False)
get_table_noise_estimates(df_OLS, df_PCA, var_names, stat_name = 'noise', 
    output_folder = output_folder, file_name = 'noise_outliers_removed_std_dev',
    show_se = True)

# "No gaps" subsample
df_OLS = collect_noise_results(var_names, '{}/no_gaps'.format(input_folder),
        model_type = 'OLS', number_lags = max_consensus_lags,
        within_variation = False)
df_PCA = collect_noise_results(var_names, '{}/no_gaps'.format(input_folder),
        model_type = 'PCA', number_factors = max_factors,
        within_variation = False)
get_table_noise_estimates(df_OLS, df_PCA, var_names, stat_name = 'noise', 
    output_folder = output_folder, file_name = 'no_gaps_noise_std_dev',
    show_se = True)

# Get stringent noise estimates
df_OLS = collect_noise_results(var_names, '{}/orthog_realiz_lag'.format(input_folder),
        model_type = 'OLS', number_lags = max_consensus_lags,
        within_variation = False)
df_PCA = collect_noise_results(var_names, '{}/orthog_realiz_lag'.format(input_folder),
        model_type = 'PCA', number_factors = max_factors,
        within_variation = False)
get_table_noise_estimates(df_OLS, df_PCA, var_names, stat_name = 'noise', 
    output_folder = output_folder, file_name = 'stringent_noise_std_dev',
    show_se = True)

df_OLS = collect_noise_results(var_names, '{}/orthog_realiz'.format(input_folder),
        model_type = 'OLS', number_lags = max_consensus_lags,
        within_variation = False)
df_PCA = collect_noise_results(var_names, '{}/orthog_realiz'.format(input_folder),
        model_type = 'PCA', number_factors = max_factors,
        within_variation = False)
get_table_noise_estimates(df_OLS, df_PCA, var_names, stat_name = 'noise', 
    output_folder = output_folder, file_name = 'stringent_realiz_only_noise_std_dev',
    show_se = True)

df_OLS = collect_noise_results(var_names, '{}/orthog_lag'.format(input_folder),
        model_type = 'OLS', number_lags = max_consensus_lags,
        within_variation = False)
df_PCA = collect_noise_results(var_names, '{}/orthog_lag'.format(input_folder),
        model_type = 'PCA', number_factors = max_factors,
        within_variation = False)
get_table_noise_estimates(df_OLS, df_PCA, var_names, stat_name = 'noise', 
    output_folder = output_folder, file_name = 'stringent_lag_only_noise_std_dev',
    show_se = True)