import datetime
import os
import sys
sys.path.append(os.path.abspath(os.path.dirname(__file__) + '/' + '../..'))

import matplotlib as mpl

mpl.use('pgf')
pgf_with_latex = {                     
    "pgf.texsystem": "xelatex",        
    "pgf.rcfonts": False,
    "text.usetex": True,                
    "font.family": "Times New Roman",
    "pgf.preamble": [
        r"\usepackage{fontspec}",    
        r"\setmainfont{Times New Roman}",        
        r"\usepackage{unicode-math}",
        r"\setmathfont{XITS Math}"
        ]
    }    
mpl.rcParams.update(pgf_with_latex)

import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
from scipy.optimize import root_scalar
import statsmodels.api as sm
import statsmodels.formula.api as smf

from python.tools import (
    clean_folder
) 

from graphs.tools import estimate_kalman_gain 

################
## Parameters ##
################

input_folder = './graphs/input'
output_folder = './graphs/output/noisy_info_calibration'
var_name = 'SPFfor_Step2'
dpi = 600
min_T = 20 # Minimum number of observations for 
           # estimating the autocorrelation of forecast errors
eps = 1e-4 # Tolerance for avoiding boundary cases

clean_folder(output_folder)

####################################
## Grab optimal number of factors ##
####################################

number_factors = pd.read_csv('{}/baseline/number_factors_GR_baseline.csv'.format(input_folder), 
                             index_col = 0)
number_factors = number_factors['Adjusted for Time F.E.'].to_dict()

##################
## Calculations ##
##################

# Read individual dataset, making sure the set of
# observations is the same as that used for the estimation
df = pd.read_csv('{}/baseline/df_resid_PCA_{}_number_factors_{}.csv'.format(input_folder,
                                                                            var_name,
                                                                            number_factors[var_name]))

# Calculate derived variables
df['E_{}_LAG1'.format(var_name)] = df.groupby('ID').shift()['E_{}'.format(var_name)]
df_cons = df.groupby('DATE').mean()
df_cons['Realiz1_LAG1'] = df_cons['Realiz1'].shift()

# Calculate the persistence of inflation
# and variance of shocks to inflation
mod = smf.ols('Realiz1 ~ Realiz1_LAG1', data = df_cons).fit()
rho = mod.params['Realiz1_LAG1']
sigma_eps = mod.resid.std()

# Estimate the autocorrelation of
# forecast errors at the consensus level
# (just for reference)
mod = smf.ols('E_{var_name} ~ E_{var_name}_LAG1'.format(var_name = var_name), data = df_cons).fit()
G = 1 - mod.params['E_{}_LAG1'.format(var_name)] / rho
res_dict = {'G_consensus': G}

# Read in noise
df_noise = pd.read_csv('{}/baseline/PCA_{}_number_factors_{}.csv'.format(input_folder, 
                                                                         'SPFfor_Step2', 
                                                                         number_factors['SPFfor_Step2']))

# Estimate persistence of forecast errors
autocorr_fe = []
for ID in df['ID'].unique():
    mask = df['ID'] == ID
    mod = smf.ols('E_{var_name} ~ E_{var_name}_LAG1'.format(var_name = var_name), 
                  data = df.loc[mask, ]).fit()
    autocorr_fe.append({'ID': ID,
                        'autocorr_fe': mod.params['E_{}_LAG1'.format(var_name)],
                        'N_autocorr': mod.nobs})
autocorr_fe = pd.DataFrame(autocorr_fe)

# Merge in to the noise dataframe
df_noise = pd.merge(df_noise, autocorr_fe, on = 'ID', how = 'left')

# Merge in variance of forecast errors
df_temp = df.groupby('ID').var()['E_{}'.format(var_name)].reset_index()
df_temp.rename(columns = {'E_{}'.format(var_name): 'var_forecast_errors'},
               inplace = True)
df_noise = pd.merge(df_noise, df_temp, on = 'ID', how = 'left')

# Only at forecasters where we have enough data
# to estimate the autocorrelation
mask = df_noise['N_autocorr'] >= min_T
df_noise = df_noise.loc[mask, ]

# Calculate implied estimates of Kalman gain
frac_to_meas_error = np.linspace(eps, 1 - eps, 100)
res = []
for alpha in frac_to_meas_error:
    res.append(estimate_kalman_gain(df_noise = df_noise, 
        alpha = alpha, 
        sigma_eps = sigma_eps, 
        rho = rho))
res = pd.DataFrame(res)
mask = res['alpha'] == eps
res_dict['G_standard_no_meas_error'] = res.loc[mask, 'G_standard'].values[0]
res_dict['G_noise_no_meas_error'] = res.loc[mask, 'G_noise'].values[0]

mask = res['alpha'] == 1 - eps
res_dict['G_standard_only_meas_error'] = res.loc[mask, 'G_standard'].values[0]

# Calculate how much measurement error
# is need to exactly match the Kalman gains
obj_fun = lambda alpha: (estimate_kalman_gain(df_noise = df_noise, 
    alpha = alpha, 
    sigma_eps = sigma_eps, rho = rho)['G_standard'] -
    estimate_kalman_gain(df_noise = df_noise, 
    alpha = alpha, 
    sigma_eps = sigma_eps, rho = rho)['G_noise'])
meas_error_necessary = root_scalar(obj_fun, method = 'brentq', bracket = (eps, 1 - eps))
res_dict['meas_error_necessary'] = meas_error_necessary.root
res_dict['G_matching'] = estimate_kalman_gain(df_noise = df_noise, 
        alpha = res_dict['meas_error_necessary'] , 
        sigma_eps = sigma_eps, 
        rho = rho)['G_standard']
res_dict['signal_to_noise'] = estimate_kalman_gain(df_noise = df_noise, 
        alpha = res_dict['meas_error_necessary'] , 
        sigma_eps = sigma_eps, 
        rho = rho)['signal_to_noise']
pd.DataFrame([res_dict]).to_csv('{}/output.csv'.format(output_folder), index = False)

# Get graph
fig, ax = plt.subplots(figsize=(4.5, 3.0))
plt.plot(res['alpha'], res['G_noise'], 
         '-k',
         label = 'Noise-Based Estimate')
plt.plot(res['alpha'], res['G_standard'], 
         '--k',
         label = 'Standard Estimate')
plt.ylabel('Kalman Gain', fontsize = 12)
plt.xlabel('Fraction of Noise Due to Measurement Error', fontsize = 12)
plt.legend(fontsize = 12, frameon = False)
plt.tight_layout()
plt.savefig("{}/implied_G.png".format(output_folder), dpi = dpi)
plt.savefig("{}/implied_G.pgf".format(output_folder), dpi = dpi)
plt.show()