import numpy as np
import pandas as pd
import statsmodels.formula.api as smf

from python.tools import (
    number_factors_GR,
    number_factors_EV
)

def collect_noise_results(var_names, input_folder,
        model_type, number_lags = None, 
        number_factors = None, within_variation = True):
    """Collect estimates of noise from
    output files
    
    Args:
        var_names (list): Variables for which noise levels are estimated
        input_folder (str): Input folder location
        model_type (str): Type of model
        number_lags (int, optional): Maximum number of consensus lags (for OLS model)
        number_factors (int, optional): Maximum number of factors (for PCA model)
        within_variation (bool, optional): If true,
            get noise as a fraction of within-forecaster variation;
            if False, get standard deviation of noise

    Returns:
        df: Estimates of noise
    """
    res = []
    for var in var_names:
        if model_type == 'OLS':
            start = 0
            end = number_lags
        elif model_type == 'PCA':
            start = 1
            end = number_factors
        else:
            raise ValueError('No such model known.')
        for ii in range(start, end + 1):
            if model_type == 'OLS':
                df = pd.read_csv('{}/OLS_{}_no_lags_{}.csv'.format(input_folder, var, ii))
            elif model_type == 'PCA':
                df = pd.read_csv('{}/PCA_{}_number_factors_{}.csv'.format(input_folder, var, ii))
            if within_variation:
                df['noise'] = 1 - df['R2']
            else:
                df['noise'] = df['sigma_v']
            mod = smf.ols('noise ~ 1', data = df).fit(cov_type = 'HC3')
            temp = {'noise': df['noise'].mean(),
                    'noise_se': mod.bse['Intercept'],
                    'autocorr_noise': df['autocorr_noise'].mean(),
                    'var': var}
            if model_type == 'OLS':
                temp['K'] = ii
            elif model_type == 'PCA':
                temp['R'] = ii
            res.append(temp)
    return pd.DataFrame(res)


def get_noise_cross_sectional_dispersion(var_names, input_folder,
        model_type, number_lags = None, 
        number_factors = None):
    """Collect results for the contribution
    of noise to cross-sectional dispersion
    
    Args:
        var_names (list): Variables for which noise levels are estimated
        input_folder (str): Input folder location
        model_type (str): Type of model
        number_lags (int, optional): Maximum number of consensus lags (for OLS model)
        number_factors (int, optional): Maximum number of factors (for PCA model)
    
    Returns:
        df: Estimates of noise
    """
    res = []
    for var in var_names:
        if model_type == 'OLS':
            start = 0
            end = number_lags
        elif model_type == 'PCA':
            start = 1
            end = number_factors
        else:
            raise ValueError('No such model known.')
        for ii in range(start, end + 1):
            if model_type == 'OLS':
                df = pd.read_csv('{}/df_resid_OLS_{}_no_lags_{}.csv'.format(input_folder, var, ii))
            elif model_type == 'PCA':
                df = pd.read_csv('{}/df_resid_PCA_{}_number_factors_{}.csv'.format(input_folder, var, ii))
            
            # Calculate cross-sectional variances of forecasts and noise levels
            df_variances = df.groupby('DATE').var().loc[:, [var, '{}_resid'.format(var)]]
            ratio = (df_variances['{}_resid'.format(var)] / df_variances[var]).median()
            temp = {'ratio': ratio,
                    'var': var}
            if model_type == 'OLS':
                temp['K'] = ii
            elif model_type == 'PCA':
                temp['R'] = ii
            res.append(temp)
    return pd.DataFrame(res)


def get_table_noise_estimates(df_OLS, df_PCA, var_names, 
        stat_name, output_folder, file_name, show_se = True,
        horizons = None):
    """Get a LaTeX table of noise level
    estimates
    
    Args:
        df_OLS (df): OLS output
        df_PCA (df): PCA output
        var_names (list): Name of variables
            for which the statistic is calculated
        stat_name (str): Name of column in df's
            containing the statistic of interest
            to be shown in table (e.g., noise)
        output_folder (str): Output folder
        file_name (str): Name of output file
        show_se (bool, optional): If True, include standard
            errors in the table
        horizons (dict, optional): Dictionary containing
            forecast horizons for each variable in
            var_names
    """
    # Extract parameters
    N = 1 + df_OLS.shape[0] + df_PCA.shape[0]
    max_consensus_lags = df_OLS['K'].max()
    max_factors = df_PCA['R'].max()

    # Construct LaTeX table
    table = '\\begin{tabular}{'
    table += '{}'.format('l')
    table += '{}|'.format('c' * (4))
    table += '{}'.format('c' * (4))
    table += '} \n'
    # Create table header
    table += '\\toprule \n'
    table += ' '
    table += '& \multicolumn{4}{c|}{Single Factor (Consensus)} & \multicolumn{4}{c}{Multiple Factors} \\\\ \n'
    table += '\\midrule \n'

    for no_lags in range(0, max_consensus_lags + 1):
        table += ' & $K = {}$'.format(no_lags)
    for no_factors in range(1, max_factors + 1):
        table += ' & $R = {}$'.format(no_factors)
    table += ' \\\\ \n'
    table += '\\midrule \n'
    
    df_res = []

    for var in var_names:
        if horizons is None:
            hh = int(var[-1]) - 1
        else:
            hh = horizons[var]
        if show_se:
            table += '\multirow{{2}}{{*}}{{$h = {}$}}'.format(hh)
        else:
            table += '$h = {}$'.format(hh)

        for no_lags in range(0, max_consensus_lags + 1):
            mask = (df_OLS['var'] == var) & (df_OLS['K'] == no_lags)
            value = df_OLS.loc[mask, stat_name].values[0]
            table += ' & {:.2f} '.format(value)
            df_res.append({'h': hh,
                           'var_name': var,
                           'method': 'OLS',
                           'no_lags/no_factors': no_lags,
                           'stat': 'coef',
                           'value': value})
        for no_factors in range(1, max_factors + 1):
            mask = (df_PCA['var'] == var) & (df_PCA['R'] == no_factors)
            value = df_PCA.loc[mask, stat_name].values[0]
            table += ' & {:.2f} '.format(value)
            df_res.append({'h': hh,
                           'var_name': var,
                           'method': 'PCA',
                           'no_lags/no_factors': no_factors,
                           'stat': 'coef',
                           'value': value})
        if show_se:
            table += ' \\\\ \n'

        table += ' '
        if show_se:
            for no_lags in range(0, max_consensus_lags + 1):
                mask = (df_OLS['var'] == var) & (df_OLS['K'] == no_lags)
                value = df_OLS.loc[mask, 'noise_se'].values[0]
                table += ' & \\footnotesize{{({:.2f})}} '.format(value)
                df_res.append({'h': hh,
                               'var_name': var,
                               'method': 'OLS',
                               'no_lags/no_factors': no_lags,
                               'stat': 'se',
                               'value': value})
            for no_factors in range(1, max_factors + 1):
                mask = (df_PCA['var'] == var) & (df_PCA['R'] == no_factors)
                value = df_PCA.loc[mask, 'noise_se'].values[0]
                table += ' & \\footnotesize{{({:.2f})}} '.format(value)
                df_res.append({'h': hh,
                               'var_name': var,
                               'method': 'PCA',
                               'no_lags/no_factors': no_lags,
                               'stat': 'se',
                               'value': value})
        if hh != 4:
            if show_se:
                table += ' \\\\ \\midrule \n'
            else:
                table += ' \\\\ \n'
        else:
            table += ' \\\\ \n'
            
    table += '\\bottomrule \n'
    table += '\\end{tabular}'

    df_res = pd.DataFrame(df_res)
    df_res.to_csv('{}/{}.csv'.format(output_folder, file_name), index = False)
    with open('{}/{}.txt'.format(output_folder, file_name), 'w+') as text_file:
        text_file.write(table)


def get_kalman_gain(rho, sigma_eps, sigma_omega):
    """Calculate the Kalman gain for a state-space model
    in which target variable x_t follows an AR(1) 
    process with

    x_t = rho x_{t - 1} + eps_t, eps_t ~ N(0, sigma_eps ** 2)

    and the observation equation is 
    
    s_t = x_t + omega_t

    where omega_t ~ N(0, sigma_omega ** 2), and s_t
    is the observed signal.
    
    Args:
        rho (float): Persistence of x_t
        sigma_eps (float): Standard deviation of shocks to x_t
        sigma_omega (float): Standard deviation of shocks to s_t
    
    Returns:
        dict: Dictionary, contains:
            shocks (float): Simulated shocks (eps)
            x (float): Simulated time series
    """
    assert rho > - 1 and rho < 1, \
        'Persistence parameter should be in (-1, 1).'
    assert (sigma_eps > 0).all() and (sigma_omega > 0).all(), \
        'Standard deviations should be positive.'
        
    Sigma = 0.5 * (-(1 - rho ** 2) * sigma_omega ** 2 
                   + sigma_eps ** 2 
                   + np.sqrt(((1 - rho ** 2) * sigma_omega ** 2 - sigma_eps ** 2) ** 2
                       + 4 * sigma_eps ** 2 * sigma_omega ** 2))
    
    return Sigma / (Sigma + sigma_omega ** 2)


def estimate_kalman_gain(df_noise, alpha, sigma_eps, rho):
    """Get standard- and noise-based estimates
    of Kalman gain. 
    
    Args:
        df_noise (df): Dataframe with autocorrelations and noise estimates
        alpha (float): Fraction of var(noise) due to measurement error
        sigma_eps (float): Standard deviation of shocks to target variable
        rho (float): Persistence of target variable
    
    Returns:
        dict: Kalman gain estimates:
            G_standard: Standard Kalman gain estimate from the autocorrelaction of
                        forecast errors
            G_noise: Noise-based Kalman gain estimate
    """
    df_temp = df_noise.copy()

    # Calculate "standard estimate"
    df_temp['var_meas_error'] = alpha * df_temp['sigma_v'] ** 2
    df_temp['kappa'] = (df_temp['var_forecast_errors'] - df_temp['var_meas_error']) / df_temp['var_forecast_errors']
    df_temp['G_standard'] = 1 - df_temp['autocorr_fe'] / (rho * df_temp['kappa'])

    # For estimates outside [0, 1],
    # set them to a boundary value
    mask = df_temp['G_standard'] < 0.0
    df_temp.loc[mask, 'G_standard'] = 0.0

    mask = df_temp['G_standard'] > 1.0
    df_temp.loc[mask, 'G_standard'] = 1.0

    # Calculate "noise-based estimate"
    df_temp['var_omega'] = ((1 - alpha) * df_temp['sigma_v'] ** 2 * (1 - rho ** 2 * (1 - df_temp['G_standard']) ** 2) 
                     / (rho ** 2 * df_temp['G_standard'] ** 2))
    df_temp['G_noise'] = get_kalman_gain(rho = rho, sigma_eps = sigma_eps, sigma_omega = np.sqrt(df_temp['var_omega'])).mean()
    
    # Deal with potential division by zero
    mask = df_temp['G_standard'] == 0.0
    df_temp.loc[mask, 'G_noise'] = 0.0
    
    signal_to_noise = sigma_eps ** 2 / df_temp['var_omega'].mean()
    
    # Return
    return {'G_standard': df_temp['G_standard'].mean(),
            'G_noise': df_temp['G_noise'].mean(),
            'alpha': alpha,
            'signal_to_noise': signal_to_noise}