import os
import sys
sys.path.append(os.path.abspath(os.path.dirname(__file__) + '/' + '../..'))

import numpy as np
from numpy.testing import assert_allclose
import pandas as pd

from python.tools import (
    simulate_dataset,
    collect_data_to_numpy,
    get_factors,
    get_loadings,
    estimate_factor_model,
    OLS_single_factor
    )


def test_collect_data_1():
    df = pd.read_csv('../PGDP/input/individual_error_statistics_PGDP.csv')
    var_name = 'SPFfor_Step4'
    data = collect_data_to_numpy(df, var_name = var_name)

    # Test a particular observation
    ID_ind = 5
    ID = data['ID_names'][ID_ind]
    mask = (df['ID'] == ID)
    test = df.loc[mask, var_name].values
    actual = data['mY'][:, ID_ind]

    return assert_allclose(test, actual)


def test_collect_data_2():
    df = pd.read_csv('../PGDP/input/individual_error_statistics_PGDP.csv')
    var_name = 'SPFfor_Step4'
    data = collect_data_to_numpy(df, var_name = var_name)

    # Test a particular observation
    ID_ind = 251
    ID = data['ID_names'][ID_ind]
    mask = (df['ID'] == ID)
    test = df.loc[mask, var_name].values
    actual = data['mY'][:, ID_ind]

    return assert_allclose(test, actual)


def test_collect_data_3():
    df = pd.read_csv('../PGDP/input/individual_error_statistics_PGDP.csv')
    df['SPFfor_Step1_LAG1'] = df.groupby('ID').shift()['SPFfor_Step1']
    
    var_name = 'SPFfor_Step4'
    exog_regressors = ['SPFfor_Step1_LAG1']
    data = collect_data_to_numpy(df, var_name = var_name, exog_regressors = exog_regressors)

    # Test a particular observation
    ID_ind = 251
    ID = data['ID_names'][ID_ind]
    mask = (df['ID'] == ID)
    test = df.loc[mask, data['EXOG_names']].values
    actual = data['mX'][ID_ind, :, :]

    return assert_allclose(test, actual)


def test_collect_data_4():
    df = pd.read_csv('../PGDP/input/individual_error_statistics_PGDP.csv')
    df['SPFfor_Step1_LAG1'] = df.groupby('ID').shift()['SPFfor_Step1']
    
    var_name = 'SPFfor_Step4'
    exog_regressors = ['Realiz1', 'SPFfor_Step1_LAG1']
    data = collect_data_to_numpy(df, var_name = var_name, exog_regressors = exog_regressors)

    # Test a particular observation
    ID_ind = 3
    ID = data['ID_names'][ID_ind]
    mask = (df['ID'] == ID)
    test = df.loc[mask, data['EXOG_names']].values
    actual = data['mX'][ID_ind, :, :]

    return assert_allclose(test, actual)

    
def test_factors_1():
    # Test whether the function for calculating
    # factor loadings is correct
    T = 100
    N = 15
    number_factors = 1
    number_exog = 0
    simulated = simulate_dataset(T, N, number_factors, number_exog = number_exog)
    df = simulated['df']
    df = pd.merge(df, simulated['df_true_loadings'][['ID', 'fixed_effect']], on = 'ID')
    
    df['forecast_adj'] = df['forecast'] - df['fixed_effect']
    matrix_data = collect_data_to_numpy(df, var_name = 'forecast_adj', 
        exog_regressors = ['exog_{}'.format(ff) for ff in range(1, number_exog + 1)])
    mL = simulated['df_true_loadings'][['loading_{}'.format(ff) for ff in range(1, number_factors + 1)]].values
    check = get_factors(matrix_data['mY'], mL)

    test = np.abs(check - simulated['df_true_factors'][['factor_{}'.format(ff) for ff in range(1, number_factors + 1)]].values).max().max()

    return assert_allclose(test, 0, atol = 1e-12)


def test_loadings_1():
    # Test whether the function for calculating
    # factor loadings is correct
    T = 100
    N = 15
    number_factors = 2
    number_exog = 0
    simulated = simulate_dataset(T, N, number_factors, number_exog = number_exog)
    df = simulated['df']

    matrix_data = collect_data_to_numpy(df, var_name = 'forecast', 
        exog_regressors = ['exog_{}'.format(ff) for ff in range(1, number_exog + 1)])
    res_loadings = get_loadings(matrix_data['mY'], 
        mF = simulated['df_true_factors'][['factor_{}'.format(ff) for ff in range(1, number_factors + 1)]].values)

    test = np.abs(res_loadings['mL'] - simulated['df_true_loadings'][['loading_{}'.format(ff) for ff in range(1, number_factors + 1)]].values).max().max()

    return assert_allclose(test, 0, atol = 1e-12)


def test_loadings_2():
    # Test whether the function for calculating
    # factor loadings is correct
    T = 100
    N = 15
    number_factors = 2
    number_exog = 3
    simulated = simulate_dataset(T, N, number_factors, number_exog = number_exog)
    df = simulated['df']

    matrix_data = collect_data_to_numpy(df, var_name = 'forecast', 
        exog_regressors = ['exog_{}'.format(ff) for ff in range(1, number_exog + 1)])
    res_loadings = get_loadings(matrix_data['mY'], 
        mF = simulated['df_true_factors'][['factor_{}'.format(ff) for ff in range(1, number_factors + 1)]].values,
        mX = matrix_data['mX'])
    test = np.abs(res_loadings['mL'] - simulated['df_true_loadings'][['loading_{}'.format(ff) for ff in range(1, number_factors + 1)]].values).max().max()

    return assert_allclose(test, 0, atol = 1e-12)


def test_least_squares_FM():
    # Test whether the function for calculating
    # factor loadings is correct
    T = 100
    N = 15
    number_factors = 4
    number_exog = 3
    simulated = simulate_dataset(T, N, number_factors, number_exog = number_exog)

    res_FM = estimate_factor_model(simulated['df'], 'forecast', number_factors,
        exog_regressors = ['exog_{}'.format(ff) for ff in range(1, number_exog + 1)], sigma_init = 1.0,
        starting_vals = 1)

    return assert_allclose(res_FM['SSR'], 0, atol = 1e-12)


def test_noise_estimate():
    # Test whether the function for calculating
    # factor loadings is correct
    T = 100
    N = 500
    number_factors = 2
    number_exog = 1
    sigma_noise = 0.5
    np.random.seed(20192019)

    simulated = simulate_dataset(T, N, number_factors, number_exog = number_exog,
        sigma_noise = sigma_noise)

    res_FM = estimate_factor_model(simulated['df'], 'forecast', number_factors,
        exog_regressors = ['exog_{}'.format(ff) for ff in range(1, number_exog + 1)], sigma_init = 1.0,
        starting_vals = 1)

    test = res_FM['res']['sigma_v'].mean() - sigma_noise

    return assert_allclose(test, 0, atol = 0.05) # Because of finite-sample, does not coincide exactly


def test_SSR():
    # Test whether SSR is calculated correctly
    T = 100
    N = 500
    number_factors = 2
    number_exog = 1
    sigma_noise = 0.5
    np.random.seed(20192019)

    simulated = simulate_dataset(T, N, number_factors, number_exog = number_exog,
        sigma_noise = sigma_noise)

    res_FM = estimate_factor_model(simulated['df'], 'forecast', number_factors,
        exog_regressors = [], sigma_init = 1.0,
        starting_vals = 1)

    test = (res_FM['df']['forecast_resid'] ** 2).mean()
    return assert_allclose(test, res_FM['SSR'])


def test_loadings():
    # Test whether loadings calculated manually
    # and via statsmodels coincide
    T = 100
    N = 500
    number_factors = 2
    number_exog = 1
    sigma_noise = 0.5
    np.random.seed(20192019)

    simulated = simulate_dataset(T, N, number_factors, number_exog = number_exog,
        sigma_noise = sigma_noise)

    res_FM = estimate_factor_model(simulated['df'], 'forecast', number_factors,
        exog_regressors = [], sigma_init = 1.0,
        starting_vals = 1)

    df_test = res_FM['df_loadings']
    df_test = pd.merge(df_test, res_FM['res'].loc[:, ['ID', 'lambda_factor_2']])
    return assert_allclose(df_test['loading_2'], df_test['lambda_factor_2'])


def test_OLS_single_factor_1():
    # Test whether the function for calculating
    # factor loadings is correct
    T = 100
    N = 15
    number_factors = 1
    number_exog = 0
    simulated = simulate_dataset(T, N, number_factors, number_exog = number_exog)

    df = simulated['df']
    df_temp = df.groupby('DATE').mean().reset_index().loc[:, ['DATE', 'forecast']]
    df_temp.rename(columns = {'forecast': 'forecast_Consensus'},
                   inplace = True)
    df = pd.merge(df, df_temp, on = 'DATE', how = 'left')

    res_OLS = OLS_single_factor(df, 'forecast', number_lags = 0)

    return assert_allclose(res_OLS['SSR'], 0, atol = 1e-12)


def test_OLS_single_factor_2():
    # Test whether the function for calculating
    # factor loadings is correct
    T = 100
    N = 15
    number_factors = 1
    number_exog = 2
    simulated = simulate_dataset(T, N, number_factors, number_exog = number_exog)

    df = simulated['df']
    df_temp = df.groupby('DATE').mean().reset_index().loc[:, ['DATE', 'forecast']]
    df_temp.rename(columns = {'forecast': 'forecast_Consensus'},
                   inplace = True)
    df = pd.merge(df, df_temp, on = 'DATE', how = 'left')

    res_OLS = OLS_single_factor(df, 'forecast', 
        exog_regressors = ['exog_{}'.format(ff) for ff in range(1, number_exog + 1)], 
        number_lags = 0)

    return assert_allclose(res_OLS['SSR'], 0, atol = 1e-12)


def test_OLS_single_factor_noise():
    # Test whether the function for calculating
    # factor loadings is correct
    T = 100
    N = 500
    number_factors = 1
    number_exog = 1
    sigma_noise = 0.5
    np.random.seed(20192019)

    simulated = simulate_dataset(T, N, number_factors, number_exog = number_exog, sigma_noise = sigma_noise)

    df = simulated['df']
    df_temp = df.groupby('DATE').mean().reset_index().loc[:, ['DATE', 'forecast']]
    df_temp.rename(columns = {'forecast': 'forecast_Consensus'},
                   inplace = True)
    df = pd.merge(df, df_temp, on = 'DATE', how = 'left')

    res_OLS = OLS_single_factor(df, 'forecast', 
        exog_regressors = ['exog_{}'.format(ff) for ff in range(1, number_exog + 1)], 
        number_lags = 0)

    test = res_OLS['res']['sigma_v'].mean() - sigma_noise
    print res_OLS['res']['sigma_v'].mean()

    return assert_allclose(test, 0, atol = 0.05)