import os
import shutil
import re

import numpy as np
import pandas as pd
import statsmodels.formula.api as smf


def get_inflation_forecasts(df, melt_id_vars):
    """Calculate inflation forecasts from
    individual-level forecasts of GDP price
    deflator.
    
    Args:
        df (df): Dataframe containing individual forecasts
        melt_id_vars (str or list): Variables for converting
            into long format when calculating dates for which
            the forecast was made
    
    Returns:
        df: Dataframe containing inflation forecasts
    """

    # Calculate annualized quarter-on-quarter inflation rates
    # Here, PGDP1 is forecast for previous quarter; PGDP2, PGDP3, ..., PGDP6
    # are forecasts for quarter t, quarter t + 1, ..., quarter t + 4 from now 
    # (see pp. 20-22 of Survey of Professional Forecasters Documentation);
    # PGDPA and PGDPB annual fourth-quarter over fourth-quarter
    # forecasts with A and B for year t and year t + 1
    # (see p. 20 of Survey of Professional Forecasters Documentation)
    for ii in range(2, 6 + 1):
        df['INF{}'.format(ii)] = 100 * ((df['PGDP{}'.format(ii)] / df['PGDP{}'.format(ii - 1)]) ** 4 - 1)

    # Melt the data into long format
    value_vars_list = ['INF{}'.format(ii) for ii in range(2, 6 + 1)]
    df = pd.melt(df, id_vars = melt_id_vars, value_vars = value_vars_list)

    # Calculate dates *for which* the forecast was made
    df['quarters_ahead'] = df['variable'].map(lambda s : int(s[-1]) - 2)
    df['DATE'] = df['DATE'] + df['quarters_ahead'].apply(lambda x: pd.DateOffset(months = x * 3))

    # Pivot back into wide format
    df = df.pivot_table(index = melt_id_vars, columns = 'variable', values = 'value')
    df = df.reset_index()
    df['DATE'] = df['DATE'].apply(lambda x: str(x.year)) + ':0' + df['DATE'].apply(lambda x: str(x.quarter))

    # Rename variables in accordance to those in SPF
    # forecast error statistics: 
    # SPFfor_Step1 means forecast for current quarter,
    # SPFfor_Step2 -- quarter t + 1, ..., SPFfor_Step5 -- quarter t + 4
    for var_name in df.columns:
        if 'INF' in var_name:
            tt = int(var_name[-1])
            new_var_name = 'SPFfor_Step{}'.format(tt - 1)
            df = df.rename(columns={var_name: new_var_name})

    return df


def clean_folder(folder):
    """Create a new folder, or if the folder already exists,
    delete all containing files
    
    Args:
        folder (string): Path to folder
    """
    if os.path.isdir(folder):
        shutil.rmtree(folder)
    try:
        os.makedirs(folder)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise


def construct_leads_lags(df, K, freq = 1, stub = '', var_names = None, group_ID = None):
    """Construct K leads and lags for
    variables in dataframe df
    
    Args:
        df (dataframe): Initial dataframe
        K (int): Number of lags and leads
        freq (int, optional): Frequency for taking lags (e.g., freq = 4
            for taking yearly lags with quarterly data)
        stub (str, optional): Optional stub added at the end 
            of each generated lead/lag variable
        var_names (list, optional): Variables for which
            to calculate leads and lags
        group_ID (None, optional): Variable
            by which to group when calculating leads/lags
            (e.g., forecaster ID)
    
    Returns:
        dataframe: Dataframe with lags
            and leads
    """
    res = df.copy()
    if var_names is None:
        var_names = list(df.columns.values)
    if group_ID is not None:
        var_names = [var for var in var_names if var != group_ID]
    for var in var_names:
        for ll in range(1, K + 1):
            var_name_lag = var + stub + '_LAG{}'.format(ll)
            var_name_lead = var + stub + '_LEAD{}'.format(ll)
            if group_ID is not None:
                res[var_name_lag] = res.loc[:, [var, group_ID]].groupby(group_ID).shift(ll * freq).loc[:, var]
                res[var_name_lead] = res.loc[:, [var, group_ID]].groupby(group_ID).shift(-ll * freq).loc[:, var]
            else:
                res[var_name_lag] = res[var].shift(ll * freq)
                res[var_name_lead] = res[var].shift(-ll * freq)
    return res


def get_indiv_factor_statistics(df, var_name, formula, Newey_West_lags = 4):
    """Convenience function for calculating
    statistics from individual-level
    time series regressions
    
    Args:
        df (df): Dataframe containing forecasts and factors_vars
        var_name (str): Variable name
        formula (str): Formula for the time-series regression
        Newey_West_lags (int, optional): Number of lags
            included when calculating Newey-West standard errors 
    
    Returns:
        df: Calculated statistics
    """
    df_temp = df.copy()

    res = []
    for ID in df['ID'].unique():
        mask = df['ID'] == ID
        df_ind = df_temp.loc[mask, ]
        std_dev_dep_var = df_ind[var_name].std()
        mod = smf.ols(formula, 
                      data = df_ind).fit(cov_type = 'HAC',
                                         cov_kwds = {'maxlags': Newey_West_lags})

        # Calculate the autocorrelation of noise
        df_temp.loc[mask, '{}_resid'.format(var_name)] = mod.resid
        df_ind = pd.DataFrame(df_temp.loc[mask, '{}_resid'.format(var_name)])
        df_ind['{}_resid_LAG1'.format(var_name)] = df_ind['{}_resid'.format(var_name)].shift()
        autocorr_noise = df_ind.corr().values[0, 1]
        sigma_v = mod.resid.std()
        R2 = mod.rsquared

        # Sometimes statsmodels does not handle perfect collinearity
        # well; in these case, replace R2 with 1.0 and sigma_v with 0.0
        if not np.isfinite(mod.rsquared) and np.abs(mod.resid).max() <= 1e-8:
            R2 = 1.0
            sigma_v = 0.0
            print('WARNING: Multicollinearity found.')

        temp = {'sigma_v': sigma_v,
                'std_dev_dep_var': std_dev_dep_var,
                'ID': ID,
                'N': mod.nobs,
                'R2': R2,
                'autocorr_noise': autocorr_noise}

        # Extract factor loadings
        for var in mod.params.index:
            if var != 'Intercept':
                temp['lambda_{}'.format(var)] = mod.params[var]
                temp['lambda_{}_se'.format(var)] = mod.bse[var]
        res.append(temp)
    res = pd.DataFrame(res)

    return {'res': res,
            'df': df_temp}


def OLS_single_factor(df, var_name, exog_regressors = [], 
        number_lags = 0, Newey_West_lags = 4):
    """Estimate noise using OLS regressions of
    
    F_{it} = a_{i0} + b_{i0} F_t + ... + b_{iK} F_{t - K}
        + exog_regressors + u_{it}
    
    with K denoting the number of lagged values
    of the consensus forecast, F_t, included
    in the regression, and exog_regressors
    denotes other controls included in the regression.

    The code assumes that the consensus forecasts
    are stored in df with the name

    var_name_Consensus

    where var_name is the name of the forecasts.

    Args:
        df (dataframe): Dataframe containing the forecasts
        var_name (str): Name of the variable
        exog_regressors (list, optional): List with names of exogenous regressors
        number_lags (int, optional): Number of lags of the
            consensus forecast to be included
        Newey_West_lags (int, optional): Number of lags
            included when calculating Newey-West standard errors
    """
    assert isinstance(number_lags, int) and number_lags >= 0, \
        'number_lags must be an integer that is >= 0'
    assert isinstance(Newey_West_lags, int) and Newey_West_lags >= 1, \
        'Newey_West_lags must be an integer that is >= 1'

    res = []
    df_temp = df.copy()
    if number_lags > 0:
        df_temp = construct_leads_lags(df, number_lags, 
                                       var_names = ['{}_Consensus'.format(var_name)],
                                       group_ID = 'ID')

    # Construct formula
    formula = '{var_name} ~ {var_name}_Consensus'.format(var_name = var_name)
    for ll in range(1, number_lags + 1):
        formula += ' + {}_Consensus_LAG{}'.format(var_name, ll)
    for regressor in exog_regressors:
        formula += ' + {}'.format(regressor)

    for ID in df_temp['ID'].unique():
        # Select data for relevant ID
        mask = df_temp['ID'] == ID
        df_ind = df_temp.loc[mask, ]
    
        # Estimate model and collect restults
        mod = smf.ols(formula, 
                      data = df_ind).fit(cov_type = 'HAC',
                                         cov_kwds = {'maxlags': Newey_West_lags})
        df_temp.loc[mask, '{}_resid'.format(var_name)] = mod.resid

    # Calculate statistics
    SSR = (df_temp['{}_resid'.format(var_name)] ** 2).mean()
    res = get_indiv_factor_statistics(df_temp, var_name, formula, 
        Newey_West_lags = Newey_West_lags)
    return {'df': res['df'],
            'res': res['res'],
            'SSR': SSR}


def predictability_regression(df, var_name, realiz_name, 
        model_type, Newey_West_lags = 4, number_factors = None, 
        number_lags = None):
    """Wrapper for a test of predictability
    of the realization by the estimated
    noise
    
    Args:
        df (dataframe): Dataframe containing the forecasts
        var_name (str): Name of variable
        realiz_name (string): Name of realization
        model_type (str): Name of model (OLS or PCA)
        Newey_West_lags (int, optional): Number of lags
            included when calculating Newey-West standard errors
        number_factors (int, optional): Number of factors
    number_lags (int, optional): Number of lags in the OLS
        consensus regressions
    
    Returns:
        df: Dataframe with results from predictability regressions
    """
    res = []
    noise_name = '{}_resid'.format(var_name)

    # Calculate "bread part" of covariance matrix
    # for adjusted standard errors
    df_time_series = df.groupby('DATE').first().reset_index()  

    for ID in df['ID'].unique():
        mask = (df['ID'] == ID) & ~(np.isnan(df[var_name]))
        df_temp = df.loc[mask, ]
        mod = smf.ols('{} ~ {}'.format(realiz_name, noise_name), 
                      data = df_temp).fit(cov_type = 'HAC',
                                          cov_kwds = {'maxlags': Newey_West_lags})
        se = get_se_predictability(df_temp, 
                           var_name = var_name, 
                           realiz_name = realiz_name, 
                           model_type = model_type,
                           number_factors = number_factors,
                           number_lags = number_lags,
                           b = None)
        temp = {'gamma': mod.params[noise_name],
                'se_unadj': mod.bse[noise_name],
                'se_adj_HC': se['se_HC'],
                'se_adj_NW': se['se_NW'],
                'ID': ID,
                'N': mod.nobs,
                'R2': mod.rsquared}
        res.append(temp)
    res = pd.DataFrame(res)
    return res


def initialize_factors(T, number_factors,
        sigma_init = 1.0):
    """Initialize factors by simulating
    random normal variables
    
    Args:
        df (df): Dataframe containing original data
        var_name (str): Name of variable
        number_factors (int): Number of factors
        sigma_init (float, optional): Standard deviation of factors when creating initial values
    
    Returns:
        df: Simulated factors
    """
    # Generate random initial values
    mF = np.zeros((T, number_factors)) * np.nan
    for ff in range(number_factors):
            mF[:, ff] = np.random.normal(loc = 0, 
                                         scale = sigma_init,
                                         size = T)
    return mF


def get_formula_time_series(var_name, number_factors,
        exog_regressors = []):
    """Construct the formula for
    the least-squares time series regression
    for estimating factor loadings
    
    Args:
        var_name (str): Variable name
        number_factors (int): Number of factors
        exog_regressors (list, optional): List with names of exogenous regressors
    
    Returns:
        str: Formula for statsmodels
    """
    formula = '{} ~ factor_1'.format(var_name)
    for ff in range(2, number_factors + 1):
        formula += ' + factor_{}'.format(ff)
    for regressor in exog_regressors:
        formula += ' + {}'.format(regressor)
    return formula


def simulate_dataset(T, N, number_factors, number_exog = 0,
        sigma_noise = 0.0):
    """Simulate a dataset for testing purposes
    
    Args:
        T (int): Time series dimension of panel
        N (int): Cross-sectional dimension of panel
        number_factors (int): Number of factors
        number_exog (int, optional): Number of exogenous variables
        sigma_noise (float, optional): Std. deviation of noise
    
    Returns:
        dict, contains: df (df): Simulated dataset
            df_true_loadings (df): True loadings
            df_true_factors (df): True factors
    """
    # Construct true factors
    df_true_factors = pd.DataFrame(np.random.normal(loc = 0.0, 
                                       scale = 1.0, 
                                       size = (T, number_factors)))
    for ii in range(number_factors):
        df_true_factors.rename(columns = {ii : 'factor_{}'.format(ii + 1)},
                               inplace = True)
    df_true_factors['DATE'] = df_true_factors.index

    # Construct exogenous variables
    if number_exog > 0:
        df_exog = pd.DataFrame(np.random.normal(loc = 0.0, 
                                       scale = 1.0, 
                                       size = (T, number_exog)))
        for ii in range(number_exog):
            df_exog.rename(columns = {ii : 'exog_{}'.format(ii + 1)},
                                   inplace = True)
        df_exog['DATE'] = df_exog.index
    else:
        df_exog = np.nan

    # Construct true loadings
    # The true loadings have a mean of 1
    df_true_loadings = pd.DataFrame(np.random.normal(loc = 1.0, 
                                       scale = 1.0, 
                                       size = (N, number_factors + 1)))

    df_true_loadings.rename(columns = {0: 'fixed_effect'},
                            inplace = True)
    for ii in range(1, number_factors + 1):
        df_true_loadings.rename(columns = {ii : 'loading_{}'.format(ii)},
                               inplace = True)
    df_true_loadings['ID'] = df_true_loadings.index
    if number_exog > 0:
        for exog in df_exog.columns:
            if exog != 'DATE':
                df_true_loadings['loading_{}'.format(exog)] = np.random.normal(loc = 0.0, 
                                       scale = 1.0, 
                                       size = (N, 1))

    df = pd.DataFrame()
    for ind in df_true_loadings['ID']:
        mask = (df_true_loadings['ID'] == ind)
        df_temp = df_true_loadings.loc[mask, ].copy()
        forecasts = np.zeros(T) + df_temp['fixed_effect'].values
        for ff in range(1, number_factors + 1):
            forecasts += df_temp['loading_{}'.format(ff)].values * df_true_factors['factor_{}'.format(ff)].values
        if number_exog > 0:
            for exog in df_exog.columns:
                if exog != 'DATE':
                    forecasts += df_temp['loading_{}'.format(exog)].values * df_exog[exog].values
        forecasts += sigma_noise * np.random.normal(loc = 0.0, 
                                                    scale = 1.0, 
                                                    size = T)
        df[ind] = forecasts

        df['DATE'] = df.index
    df = pd.melt(df, id_vars = 'DATE')
    df.rename(columns = {'variable': 'ID',
                         'value': 'forecast'},
                   inplace = True)

    if number_exog > 0:
        df = pd.merge(df, df_exog, on = 'DATE', how = 'left')

    return {'df': df,
            'df_true_loadings': df_true_loadings,
            'df_true_factors': df_true_factors}


def summary_statistics_table(df, mask, var_names, output_name):
    """Convenience function for providing the
    summary statistics table. The function is hard coded
    and does not work on generic infputs
    
    Args:
        df (df): Dataset
        mask (bool): Boolean for the restricted sample
        var_names (list): Variable names
        output_name (str): Output file name
    
    Returns:
        str: LaTeX table
    """
    df_temp = df.copy()
    res = []
    res_restricted = []

    # Get forecasts adjusted for forecaster
    # fixed effects
    df_temp = adjust_forecaster_effects(df_temp, var_names)

    # Calculate number of forecasts made by each forecaster
    # and the cross-sectional dimension for each date
    for var in var_names:
        df_temp = get_mask_minimum_obs(df_temp, var, 
            min_N = 1, min_T = 1,
            return_df = True)['df']

    # Calculate summary statistics
    for var in var_names:
        res.append({'Total Forecasts': df_temp[var].count(),
                    'Med. \# Forecasts / Forecaster': df_temp['N_ID_{}'.format(var)].median(),
                    'Med. \# Forecasts / Survey': df_temp['N_DATE_{}'.format(var)].median(),
                    'Avg. Forecast': df_temp[var].mean(),
                    'Min. Forecast': df_temp[var].min(),
                    'Max. Forecast': df_temp[var].max(),
                    'Avg. Forecast Error': df_temp['E_{}'.format(var)].mean(),
                    'Std. Dev. Forecast': df_temp[var].std(),
                    'Std. Dev. Forecast, Within': df_temp['{}_fe_adj'.format(var)].std()})
        res_restricted.append({'Total Forecasts': df_temp.loc[mask, var].count(),
                               'Avg. Forecast': df_temp.loc[mask, var].mean(),
                               'Min. Forecast': df_temp.loc[mask, var].min(),
                               'Max. Forecast': df_temp.loc[mask, var].max(),
                               'Avg. Forecast Error': df_temp.loc[mask, 'E_{}'.format(var)].mean(),
                               'Std. Dev. Forecast': df_temp.loc[mask, var].std(),
                               'Std. Dev. Forecast, Within': df_temp.loc[mask, '{}_fe_adj'.format(var)].std()})
    res = pd.DataFrame(res)
    res_restricted = pd.DataFrame(res_restricted)

    # Re-order the columns
    res = res[['Total Forecasts',
               'Med. \# Forecasts / Forecaster',
               'Med. \# Forecasts / Survey',
               'Avg. Forecast',
               'Min. Forecast',
               'Max. Forecast',
               'Avg. Forecast Error',
               'Std. Dev. Forecast',
               'Std. Dev. Forecast, Within']]
    res_restricted = res_restricted[['Total Forecasts',
                                     'Avg. Forecast',
                                     'Min. Forecast',
                                     'Max. Forecast',
                                     'Avg. Forecast Error',
                                     'Std. Dev. Forecast',
                                     'Std. Dev. Forecast, Within']]

    # Format the table
    res['Total Forecasts'] = res['Total Forecasts'].apply(lambda x: '{:,}'.format(x))
    res_restricted['Total Forecasts'] = res_restricted['Total Forecasts'].apply(lambda x: '{:,}'.format(x))

    for var in ['Med. \# Forecasts / Forecaster',
                'Med. \# Forecasts / Survey',
                'Avg. Forecast',
                'Min. Forecast',
                'Max. Forecast',
                'Avg. Forecast Error',
                'Std. Dev. Forecast',
                'Std. Dev. Forecast, Within']:
        res[var] = res[var].apply(lambda x: '{:.2f}'.format(x))

    for var in ['Avg. Forecast',
                'Min. Forecast',
                'Max. Forecast',
                'Avg. Forecast Error',
                'Std. Dev. Forecast',
                'Std. Dev. Forecast, Within']:    
        res_restricted[var] = res_restricted[var].apply(lambda x: '{:.2f}'.format(x))

    res = res.T
    for col in res.columns:
        res.rename(columns = {col : '$h = {}$'.format(col)},
                   inplace = True)

    latex = res.to_latex(escape = False)
    latex = latex.replace('\\midrule', '\\midrule & \\multicolumn{5}{c}{Initial Sample} \\\\ \\midrule')

    # Append statistics from the
    # restricted sample
    latex_restr = res_restricted.T.to_latex()
    latex_restr = re.search(r'\\midrule\n.*?\n\\bottomrule', latex_restr, re.DOTALL).group()
    latex_restr = str(latex_restr).replace('\\midrule', '\\midrule & \\multicolumn{5}{c}{Final Sample} \\\\ \\midrule')
    latex = latex.replace('\\bottomrule', latex_restr)

    with open(output_name, 'w+') as text_file:
        text_file.write(latex)


def get_var_names_predictability(var_name, realiz_name, model_type,
        number_factors = None, number_lags = None):
    """Convenience function for getting all the
    names that are necessary for calculating
    standard errors corrected for generated
    regressors

    
    Args:
        var_name (str): Name of dependent variable
        realiz_name (str): Name of realization
        model_type (str): Name of model (OLS or PCA)
        number_factors (int, optional): Number of factors
        number_lags (int, optional): Number of lags in the OLS
        consensus regressions
    
    Returns:
        tuple: (variable names of variables in z_{i,t},
                all variables that are necessary for calculating covariance matrix,
                names of factors)
    
    """
    var_names_z = [realiz_name, 'intercept']

    # Get names of factors
    if model_type == 'OLS':
        factor_names = ['{}_Consensus'.format(var_name)]
        for ll in range(1, number_lags + 1):
            factor_names.append('{}_Consensus_LAG{}'.format(var_name, ll))
    elif model_type == 'PCA':
        factor_names = ['factor_{}'.format(ff) for ff in range(1, number_factors + 1)]
    else:
        raise ValueError('No such model known.')
    
    # Append factors
    for factor in factor_names:
        var_names_z.append(factor)
    
    # Get another list with noise appended
    var_names_all = list(var_names_z)
    var_names_all.append('{}_resid'.format(var_name))

    return (var_names_z, var_names_all, factor_names)


def calculate_bread_part(df, var_name, realiz_name, model_type,
        number_factors = None, number_lags = None):
    """Calculate the "bread part" of the sandwich
    covariance matrix that accounts for the 
    generated regressors problem
    
    Args:
        df (df): Dataframe containing original data
        var_name (str): Name of variable
        realiz_name (str): Name of realization 
        model_type (str): Name of model (OLS or PCA)
        number_factors (int, optional): Number of factors
        number_lags (int, optional): Number of lags in the OLS
            consensus regressions
    
    Returns:
        array: Sandwich component in covariance matrix

    """
    df_temp = df.copy()
    df_temp['intercept'] = 1.0

    # Get variable names
    _, var_names_all, factor_names = get_var_names_predictability(var_name, 
        realiz_name, model_type,
        number_factors = number_factors, 
        number_lags = number_lags)

    # Calculate the number of factors (in PCA case)
    # or number of lags
    if model_type == 'OLS':
        R = number_lags + 1
    elif model_type == 'PCA':
        R = number_factors

    # Get relevant variables and drop NaN
    df_temp = df_temp[var_names_all].dropna()

    # Calculate bread part of covariance matrix
    mean_factor = []
    for factor in factor_names:
        mean_factor.append(df_temp[factor].mean())
    mean_factor = np.array([mean_factor]).T
    sigma_factor = df_temp[factor_names].values
    sigma_factor = np.matmul(sigma_factor.T, sigma_factor) / float(df_temp.shape[0])

    A = np.block([[1, mean_factor.T],
                  [mean_factor, sigma_factor]])

    # Calculate mean realization and mean
    # of (realization * factor_i)
    mean_realiz = df_temp['Realiz1'].mean()

    cross_prod_names = []
    for factor in factor_names:
        df_temp['{}_prod_realiz'.format(factor)] = df_temp[factor] * df_temp['Realiz1']
        cross_prod_names.append('{}_prod_realiz'.format(factor))

    mean_realiz_x_factor = df_temp.loc[:, cross_prod_names].mean().values
    mean_realiz_x_factor = np.array([mean_realiz_x_factor]).T
    z = np.block([[mean_realiz],
                  [mean_realiz_x_factor]])
    b = np.matmul(np.linalg.inv(A), z)
    b = np.block([[1],
                  [-b]])
    return b


def get_se_predictability(df, var_name, realiz_name, model_type,
        number_factors = None, number_lags = None, b = None, 
        lags_Newey_West = 4):
    """Calculate standard errors for the
    predictability regression that take into
    account generated regressors (for a given
    forecaster)
    
    Args:
        df (df): Dataframe containing original data for a single forecaster
        var_name (str): Name of variable
        realiz_name (str): Name of realization 
        model_type (str): Name of model (OLS or PCA)
        number_factors (int, optional): Number of factors
        number_lags (int, optional): Number of lags in the OLS
            consensus regressions
        b (array, optional): Bread part of the covariance matrix
        lags_Newey_West (int, optional): Number of lags included 
            in Newey-West standard errors
    
    Returns:
        float: Standar error
    """
    if df['ID'].unique().shape[0] != 1:
        raise ValueError('There should be a single ID in the data.')
    
    df_temp = df.copy()
    df_temp['intercept'] = 1.0

    # Get the "bread" part of covariance matrix
    if b is None:
        df_time_series = df_temp.groupby('DATE').first().reset_index()
        b = calculate_bread_part(df_temp, var_name, realiz_name, model_type,
                number_factors = number_factors, 
                number_lags = number_lags)
    
    # Get variable names
    var_names_z, var_names_all, factor_names = get_var_names_predictability(var_name, 
        realiz_name, model_type,
        number_factors = number_factors, 
        number_lags = number_lags)

    # Drop NaN's
    df_temp = df_temp[var_names_all]
    df_temp = df_temp.dropna().reset_index()

    # Calculate the number of factors (in PCA case)
    # or number of lags
    if model_type == 'OLS':
        R = number_lags + 1
    elif model_type == 'PCA':
        R = number_factors

    resid_matrix = np.diag(df_temp['{}_resid'.format(var_name)].values ** 2)
    Sigma = np.matmul(df_temp[var_names_z].values.T, resid_matrix)
    Sigma = np.matmul(Sigma, df_temp[var_names_z].values) / float(df_temp.shape[0])

    NW_part = np.zeros((R + 2, R + 2))
    for ll in range(1, lags_Newey_West + 1):
        weight = 1 - ll / float(lags_Newey_West + 1)
        
        Z = df_temp.iloc[ll:, ]
        Z = Z[var_names_z].values
        Z_lag = df_temp.iloc[:-ll, ]
        Z_lag = Z_lag[var_names_z].values

        resid_matrix = (df_temp['{}_resid'.format(var_name)][ll:].values 
                      * df_temp['{}_resid'.format(var_name)][:-ll].values)
        resid_matrix = np.diag(resid_matrix)
        NW_part += weight * (np.matmul(np.matmul(Z.T, resid_matrix), Z_lag) + 
                                 np.matmul(np.matmul(Z_lag.T, resid_matrix), Z))
    NW_part = NW_part / float(df_temp.shape[0])
    Sigma_NW = Sigma + NW_part
    xi = ((df_temp['{}_resid'.format(var_name)] ** 2).mean()) ** (-1)

    res = []
    for cov_mat in [Sigma, Sigma_NW]:
        se = np.sqrt(np.matmul(np.matmul(b.T, cov_mat), b) * xi ** 2 / float(df_temp.shape[0]))
        res.append(se[0][0])
    
    return {'se_HC': res[0],
            'se_NW': res[1]}


def get_OLS_results(df, var_names, realiz_name, max_consensus_lags, output_folder,
        Newey_West_lags = 4, exog_regressors = [], lagged_forecast = False):
    """Convenience wrapper for calculating
    all results for OLS estimation using
    consensus forecasts
    
    Args:
        df (df): Dataset
        var_names (list): Variables for which to estimate noise levels
        realiz_name (str): Name of realization
        max_consensus_lags (int): Maximum number of lags of consensus to include
        output_folder (str): Name of output folder for results
        Newey_West_lags (int, optional): Number of lags included 
            in Newey-West standard errors
        exog_regressors (list, optional): List with names of exogenous regressors
        lagged_forecast (bool, optional): If True, add lagged forecast as 
            an exogenous regressor
    """
    for var in var_names:
        for number_lags in range(0, max_consensus_lags + 1):
            if lagged_forecast:
                    lags = ['{}_LAG1'.format(var)]
            else:
                lags = []
            res = OLS_single_factor(df, var_name = var, number_lags = number_lags, 
                Newey_West_lags = Newey_West_lags,
                exog_regressors = exog_regressors + lags)
            if realiz_name is not None:
                pred_test = predictability_regression(df = res['df'], 
                    var_name = var, realiz_name = realiz_name, 
                    model_type = 'OLS', Newey_West_lags = Newey_West_lags, 
                    number_lags = number_lags)
                pred_test.to_csv('{}/predictability_OLS_{}_no_lags_{}.csv'.format(output_folder, var, number_lags), index = False)
            res['df'].to_csv('{}/df_resid_OLS_{}_no_lags_{}.csv'.format(output_folder, var, number_lags), index = False)
            res['res'].to_csv('{}/OLS_{}_no_lags_{}.csv'.format(output_folder, var, number_lags), index = False)


def get_PCA_results(df, var_names, realiz_name, max_factors, 
        output_folder, Newey_West_lags = 4, exog_regressors = [],
        lagged_forecast = False,
        eps = 1e-4, max_iter = 100, starting_vals = 10):
    """Convenience wrapper for calculating
    all results for PCa estimation of noise
    levels
    
    Args:
        df (df): Dataset
        var_names (list): Variables for which to estimate noise levels
        realiz_name (str): Name of realization
        max_factors (int): Maximum number of factors
        output_folder (str): Name of output folder for results
        Newey_West_lags (int, optional):  Number of lags included 
            in Newey-West standard errors
        exog_regressors (list, optional): List with names of exogenous regressors
        lagged_forecast (bool, optional): If True, add lagged forecast as 
            an exogenous regressor
        eps (float, optional): Tolerance level for convergence
        max_iter (int, optional): Maximum number of iterations
        starting_vals (int, optional): Number of starting values
    """
    res_flag = []

    res_SSR = []
    for var in var_names:
        formula = '{} ~ C(ID)'.format(var)
        for regressor in exog_regressors:
            formula += ' + {}'.format(regressor)
            if lagged_forecast:
                formula += ' + {}_LAG1'.format(var)
        mod = smf.ols(formula, data = df).fit()
        res_SSR.append({'number_factors': 0,
                        'var': var,
                        'SSR': (mod.resid ** 2).mean()})

    for var in var_names:
          for number_factors in range(1, max_factors + 1):
                if lagged_forecast:
                    lags = ['{}_LAG1'.format(var)]
                else:
                    lags = []
                res = estimate_factor_model(df, var_name = var, 
                    exog_regressors = exog_regressors + lags,
                    number_factors = number_factors,
                    eps = eps, 
                    max_iter = max_iter, 
                    starting_vals = starting_vals, 
                    Newey_West_lags = Newey_West_lags)

                if realiz_name is not None:
                    res['df'][realiz_name] = df[realiz_name].values
                    pred_test = predictability_regression(res['df'], var_name = var, 
                        realiz_name = realiz_name, model_type = 'PCA', 
                        Newey_West_lags = Newey_West_lags, number_factors = number_factors)
                    pred_test.to_csv('{}/predictability_PCA_{}_number_factors_{}.csv'.format(output_folder, var, number_factors), index = False)
                
                res['df'].to_csv('{}/df_resid_PCA_{}_number_factors_{}.csv'.format(output_folder, var, number_factors), index = False)
                res['res'].to_csv('{}/PCA_{}_number_factors_{}.csv'.format(output_folder, var, number_factors), index = False)
                res_SSR.append({'number_factors': number_factors,
                                'var': var,
                                'SSR': res['SSR']})
                res_flag.append({'number_factors': number_factors,
                                 'var': var,
                                 'flag': res['flag']})

                # Save output (repeatedly throughout the iterations)
                pd.DataFrame(res_SSR).to_csv('{}/SSR_PCA_{}.csv'.format(output_folder, var), index = False)
                pd.DataFrame(res_flag).to_csv('{}/flag_PCA_{}.csv'.format(output_folder, var), index = False)


def get_mask_minimum_obs(df, var_name, min_N, min_T,
        return_df = False, iterated_mask = True,
        max_iter = 100):
    """Get a mask for sample selection
    based on requiring (a) at least min_T
    forecasts for each forecaster; and
    (b) at least min_N forecasters for
    each date
    
    Args:
        df (df): Dataset
        var_name (str): Name of variable for which
            the requirements are based 
        min_N (int): Minimal cross-sectional dimension
        min_T (int): Minimal time-series dimension
        return_df (bool, optional): If true, return
            the dataframe with the calculated numbers of
            observations
        iterated_mask (bool, optional): If true,
            iterate to ensure that the conditions
            on min_N and min_T are satisfied simultaneously
        max_iter (int, optional): Number of iterations
            if iterated_mask = True
    
    Returns:
        mask: pandas mask of Booleans
    """
    df_temp = df.copy()

    # Calculate number of forecasts made by each forecaster
    df_count = df_temp.groupby('ID').count().loc[:, var_name]
    df_count.rename('N_ID_{}'.format(var_name), inplace = True)
    df_temp = pd.merge(df_temp, df_count, left_on = 'ID', right_index = True, how = 'left')

    # Calculate the cross-sectional dimension for each date
    df_count = df_temp.groupby('DATE').count().loc[:, var_name]
    df_count.rename('N_DATE_{}'.format(var_name), inplace = True)
    df_temp = pd.merge(df_temp, df_count, left_on = 'DATE', right_index = True, how = 'left')

    # Get mask
    mask = ((df_temp['N_DATE_{}'.format(var_name)] >= min_N) 
          & (df_temp['N_ID_{}'.format(var_name)] >= min_T))

    if iterated_mask:
        diff = np.inf
        no_iter = 0
        old_mask = mask
        while diff > 0 and no_iter < max_iter:
            df_temp_new = df.copy()
            df_temp_new.loc[~old_mask, var_name] = np.nan
            new_mask = get_mask_minimum_obs(df_temp_new, var_name,
                                            min_N, min_T, 
                                            iterated_mask = False)
            diff = np.sum(old_mask) - np.sum(new_mask)
            old_mask = new_mask
        if no_iter == max_iter:
            raise ValueError('Maximum number of iterations reached.')
        mask = new_mask

    if not return_df:
        return mask
    else:
        return {'mask': mask,
                'df': df_temp}


def adjust_forecaster_effects(df, var_names):
    """Convenience function to adjust variables
    in var_names for forecaster fixed effects
    (within transformation)
    
    Args:
        df (df): Dataset
        var_names (list): Variable names
    
    Returns:
        df: Dataset, including adjusted values
    """
    df_temp = df.copy()

    # Calculate forecasts, adjusted for 
    # forecaster fixed effects
    df_mean = df_temp.groupby('ID').mean().loc[:, var_names].reset_index()
    for var in var_names:
        df_mean.rename(columns = {var: '{}_mean'.format(var)},
                       inplace = True)
    df_temp = pd.merge(df_temp, df_mean, on = 'ID', how = 'left')
    for var in var_names:
        df_temp['{}_fe_adj'.format(var)] = (df_temp[var] - df_temp['{}_mean'.format(var)] 
                                + df_temp[var].mean())
        del df_temp['{}_mean'.format(var)]

    return df_temp


def collect_data_to_numpy(df, var_name, exog_regressors = []):
    """Convenience function for converting
    pandas dataframe to numpy matrices
    for use in further calculations
    
    Args:
        df (df): Pandas dataframe
        var_name (str): Dependent variable
        exog_regressors (list, optional): Exogenous regressors
    
    Returns:
        dict: 
            mY (np matrix): (T x N) matrix of dependent variable
            mX (np matrix): (N x T X K) matrix of exogenous variables

    """
    # Construct matrix of dependent variables
    # of dimension (T x N)
    mY = pd.pivot(df[['ID', 'DATE', var_name]], 
        index = 'DATE', 
        columns = 'ID', values = var_name)
    ID_names   = mY.columns.values
    DATE_names = mY.index.values
    mY = mY.values

    # Collect dimensions
    T = mY.shape[0]
    N = mY.shape[1]
    K = len(exog_regressors)

    # Construct array of exogenous variables
    # of dimension (N x T x K)
    # First dimension indexes forecasters, 
    # and yields an (T x K) matrix
    # with K denoting the number of exogenous variables
    if len(exog_regressors) > 0:
        mTemp = pd.pivot(df[exog_regressors + ['ID', 'DATE']], 
            index = 'DATE', columns = 'ID', 
            values = exog_regressors)
        mTemp.columns = mTemp.columns.swaplevel(0, 1)
        mTemp.sort_index(axis = 1, level = 0, inplace = True)
        mX = np.zeros((N, T, K)) * np.nan
        for ii, ID in enumerate(ID_names):
            mX[ii, :, :] = mTemp[ID][exog_regressors].values
        res = {'mX': mX}
    else:
        res = {}

    res['mY'] = mY
    res['ID_names'] = ID_names
    res['DATE_names'] = DATE_names
    res['EXOG_names'] = exog_regressors

    return res


def get_loadings(mY, mF, mX = None):
    """Estimate factor loadings
    using time series regressions
    
    Args:
        mY (np matrix): (T x N) matrix
            containing the LHS variable
        mF (np matrix): (T x R) matrix
            containing factors
        mX (np matrix, optional): (N x T x K)
            matrix containing potential
            exogenous regressors
    
    Returns:
        dict: contains:
            mRes (np matrix): (N x (R + K + 1))
                matrix containing factor loadings
                and fixed effects
            mY_adj (np_matrix): (T x N) matrix
                containing the LHS variable
                after subtracting the fixed effect
                and the effect of any exogenous
                variables
            mL (np matrix): (N x R) matrix of
                factor loadings
            SSR (float): Sum of squared residuals
    """
    assert mY.shape[0] == mF.shape[0], \
        'mY and mF must have the same number of rows'
    if mX is not None:
        assert mY.shape[0] == mX.shape[1], \
            'Dimensions do not match'

    # Extract dimensions
    T = mY.shape[0]
    N = mY.shape[1]
    R = mF.shape[1]
    if mX is not None:
        K = mX.shape[2]
    else:
        K = 0
    
    # Matrices for results
    mRes   = np.zeros((N, K + R + 1)) * np.nan
    mY_adj = np.zeros((T, N)) * np.nan
    SSR = 0
    NT = 0

    # Loop over individual units
    for ii in range(mY.shape[1]):
        # Collect all data and get non-missings
        if mX is not None:
            mData = np.concatenate((mF, mX[ii, :, :]), axis = 1).copy()
        else:
            mData = mF
        mData = np.insert(mData, 0, 1, axis = 1) # Intercept
        mData = np.insert(mData, 0, mY[:, ii], axis = 1) 
        mask = ~np.isnan(mData).any(axis=1)
        mData = mData[mask]

        # Estimate regression coefficients
        if mData[:, 0].shape[0] >= mData[:, 1:].shape[1]:
            mRes[ii, :] = np.matmul(np.linalg.pinv(mData[:, 1:]), mData[:, 0])
        else:
            mRes[ii, :] = np.nan

        # Calculate values of mY adjusted
        # for fixed effects and, if they exist, exogenous variables
        if K > 0:
            mY_adj[mask, ii] = mY[mask, ii] - mRes[ii,  0] - np.matmul(mData[:, (R + 2):], mRes[ii,  (R + 1):])
        else:
            mY_adj[mask, ii] = mY[mask, ii] - mRes[ii,  0]

        # Calculate SSR
        SSR += np.sum((mY[mask, ii] - np.matmul(mData[:, 1:], mRes[ii, :])) ** 2)
        NT += len(mY[mask, ii])
    
    return {'mRes': mRes,
            'mY_adj': mY_adj,
            'mL': mRes[:, 1:(R + 1)],
            'SSR': SSR / NT}


def get_factors(mY_adj, mL):
    """Estimate factors using
    cross-sectional regressions
    
    Args:
        mY_adj (np matrix): (T x N) matrix
            containing the LHS variable
            after subtracting the fixed effect
            and the effect of any exogenous
            variables
        mL (np matrix): (N x R) matrix of
                factor loadings
    
    Returns:
        np matrix: (T x R) matrix of factors
    """
    assert mY_adj.shape[1] == mL.shape[0], \
            'Dimensions do not match'

    # Extract dimensions
    T = mY_adj.shape[0]
    N = mY_adj.shape[1]
    R = mL.shape[1]
    
    # Loop across time periods
    mF = np.zeros((T, R)) * np.nan
    for ii in range(mY_adj.shape[0]):
        # Collect data and select non-missings
        mData = mL.copy()
        mData = np.insert(mData, 0, mY_adj[ii, ], axis = 1)
        mask = ~np.isnan(mData).any(axis=1)
        mData = mData[mask]

        # Estimate regression coefficients
        if mData[:, 0].shape[0] >= mData[:, 1:].shape[1]:
            mF[ii, :] = np.matmul(np.linalg.pinv(mData[:, 1:]), mData[:, 0])
        else:
            mF[ii, :] = np.nan

    return mF


def estimate_factor_model(df, var_name, number_factors, 
        exog_regressors = [], sigma_init = 1.0, 
        eps = 1e-4, max_iter = 100, starting_vals = 5,
        Newey_West_lags = 4):
    """Estimate factor model (FM)
    using iterated least squares
    
    Args:
        df (df): Dataframe containing original data
        var_name (str): Name of variable
        number_factors (int): Number of factors
        exog_regressors (list, optional): List with names of exogenous regressors
        sigma_init (float, optional): Standard deviation of loadings when creating initial values   
        eps (float, optional): Tolerance level for convergence
        max_iter (int, optional): Maximum number of iterations
        starting_vals (int, optional): Number of starting values
        Newey_West_lags (int, optional): Number of lags
            included when calculating Newey-West standard errors
    
    Returns:
        dict, contains:
            df (df): Dataframe, original data and estimated factors and residuals
            df_loadings (df): Estimated factor loadings
            df_factors (df): Estimated factors
            res (df): Estimated individual-level statistics
            SSR (float): Sum of squared residuals
    """
    # Get data in numpy arrays
    matrix_data = collect_data_to_numpy(df, var_name, exog_regressors = exog_regressors)
    mY = matrix_data['mY']
    if len(exog_regressors) > 0:
        mX = matrix_data['mX']
    else:
        mX = None

    # Initialize
    res = []
    SSR = []
    T = mY.shape[0]

    # Loop across starting value
    for ii in range(starting_vals):
        # Generate starting values
        mF = initialize_factors(T, number_factors, sigma_init = sigma_init)

        # Initialize
        number_iter = 0
        SSR_old = np.inf
        improvement = np.inf
        res_loadings = get_loadings(mY = mY, mF = mF, mX = mX)

        # Main estimation loop
        while number_iter < max_iter and improvement > eps:
            res_factors = get_factors(mY_adj = res_loadings['mY_adj'], 
                                     mL = res_loadings['mL'])
            mF = res_factors
            res_loadings = get_loadings(mY = mY, mF = mF, mX = mX)
            SSR_new = res_loadings['SSR']
            improvement = SSR_old - SSR_new
            SSR_old = SSR_new
            number_iter += 1

            if improvement < 0:
                print('    WARNING: Decrease in SSR of {}'.format(improvement))

        if number_iter == max_iter or improvement < 0:
            flag = 1
        else:
            flag = 0

        res.append({'mF': mF,
                    'mRes': res_loadings['mRes'],
                    'ii': ii,
                    'mL': res_loadings['mL'],
                    'flag': flag})
        SSR.append(SSR_new)
    
    # Extract estimation results with lowest SSR
    ind_min = np.argmin(np.array(SSR))
    SSR = SSR[ind_min]
    flag = res[ind_min]['flag']
    
    # Collect estimated factors
    df_factors = pd.DataFrame(res[ind_min]['mF'])
    df_factors.columns = ['factor_{}'.format(ff) for ff in range(1, number_factors + 1)]
    df_factors['DATE'] = matrix_data['DATE_names']

    # Collect estimated loadings
    df_loadings = pd.DataFrame(res[ind_min]['mRes'])
    col_names = (['fixed_effect'] 
               + ['loading_{}'.format(ff) for ff in range(1, number_factors + 1)])
    if len(exog_regressors) > 0:
        col_names += ['loading_{}'.format(ff) for ff in exog_regressors]
    df_loadings.columns = col_names
    df_loadings['ID'] = matrix_data['ID_names']

    # Merge factors to main dataframe
    df = pd.merge(df, df_factors, on = 'DATE', how = 'left')

    # Calculate individual-level statistics
    formula = get_formula_time_series(var_name, number_factors,
        exog_regressors = exog_regressors)
    stats = get_indiv_factor_statistics(df, var_name, formula, 
        Newey_West_lags = Newey_West_lags)
    
    return {'df': stats['df'],
            'df_loadings': df_loadings,
            'df_factors': df_factors,
            'res': stats['res'],
            'SSR': SSR,
            'flag': flag}


def number_factors_GR(df_SSR):
    """Calculate the optimal number
    of factors according to the SSR-based
    growth-ratio statistic of Ahn and Horenstein
    (2013)
    
    Args:
        df_SSR (df): Dataframe with SSR
            for models with different numbers of factors
    
    Returns:
        int: Optimal number of factors
    """
    df_temp = df_SSR.copy()
    df_temp = df_temp.sort_values('number_factors', ascending = True)
    df_temp['SSR_growth'] = df_temp['SSR'].shift() / df_temp['SSR']
    df_temp['GR_statistic'] = np.log(df_temp['SSR_growth']) / np.log(df_temp['SSR_growth'].shift(-1))
    mask = df_temp['GR_statistic'] == df_temp['GR_statistic'].max()
    return df_temp.loc[mask, 'number_factors'].values[0]


def number_factors_EV(df, max_number_factors):
    """Calculate the optimal number
    of factors according to the eigenvalue
    based statistic proposed by Chen,
    Fernandez-Val and Weidner (2019)
    
    Args:
        df (df): Dataframe
        max_number_factors (int): Maximum number of factors
    
    Returns:
        int: Optimal number of factors
    """
    LR_matrix = pd.pivot(df, index = 'ID', columns = 'DATE', values = 'LR_component')
    LR_matrix = LR_matrix.dropna(axis = 1).values
    eig_vals, _ = np.linalg.eigh(np.matmul(LR_matrix, LR_matrix.T))
    eig_vals = eig_vals[::-1] # Sort eigenvalues
    res = []
    for ii in range(max_number_factors):
        res.append({'EV': eig_vals[ii] / eig_vals[ii + 1],
                    'number_factors': ii + 1})
    res = pd.DataFrame(res)
    mask = res['EV'] == res['EV'].max()
    return res.loc[mask, 'number_factors'].values[0]


def get_table_number_of_factors_GR(output_folder, 
        output_str, number_factors):
    """Get a table with optimal number
    of factors according to GR statistic
    
    Args:
        output_folder (str): Output folder
        output_str (str): String added to output
        number_factors (int): Number of factors
            used for calculating the statistic
    """
    # Get number of factors according
    # to the GR statistic
    
    # Collect SSR values
    df_raw = pd.read_csv('{}/SSR_PCA_SPFfor_Step5.csv'.format(output_folder))
    df_adj = pd.read_csv('{}/SSR_PCA_SPFfor_Step5_cons_adj.csv'.format(output_folder))
    df_SSR = pd.DataFrame([])
    for df in [df_raw, df_adj]:
        for var in df['var'].unique():
            mask = (df['var'] == var)
            df_temp = df.loc[mask, ].copy().reset_index()
            df_SSR = pd.concat([df_SSR, df_temp])
    df_SSR.to_csv('{}/SSR_{}.csv'.format(output_folder, output_str),
        index = False)

    # Get number of factors
    factors_GR = df_SSR.groupby('var').aggregate(number_factors_GR)['number_factors'].reset_index()
    factors_GR['cons_adj'] = (factors_GR['var'].apply(lambda x: 'cons_adj' in x)) * 1.0
    factors_GR['horizon'] = factors_GR['var'].apply(lambda x: [int(s) for s in x if s.isdigit()][0]) - 1
    latex = pd.pivot_table(factors_GR, index = 'cons_adj', columns = 'horizon', values = 'number_factors')
    latex.columns = ['$h = {}$'.format(hh) for hh in range(0, 4 + 1)]
    latex.index = ['Raw Forecasts', 'Adjusted for Time F.E.']
    del latex.columns.name, latex.index.name

        # Export output to CSV for later use
    df_output = latex.T.copy()
    df_output.index = ['SPFfor_Step{}'.format(hh) for hh in range(1, 5 + 1)]
    df_output.to_csv('{}/number_factors_GR_{}.csv'.format(output_folder, output_str))

    # Export to LaTeX
    latex = latex.to_latex(escape = False)
    with open('{}/number_factors_GR_{}.txt'.format(output_folder, output_str), 'w+') as text_file:
        text_file.write(latex)


def get_table_number_of_factors_EV(output_folder, 
        output_str, number_factors):
    """Get a table with optimal number
    of factors according to EV statistic
    
    Args:
        output_folder (str): Output folder
        output_str (str): String added to output
        number_factors (int): Number of factors
            used for calculating the statistic
    """
    res_EV = []

    for ii in range(1, 5 + 1):
        for cons_adj in [0, 1]:
            var_name = 'SPFfor_Step{}'.format(ii)
            if cons_adj == 1:
                var_name += '_cons_adj'
                
            # Read in data
            df_res = pd.read_csv('{}/df_resid_PCA_{}_number_factors_{}.csv'.format(output_folder, var_name, number_factors))
            df_loadings = pd.read_csv('{}/PCA_{}_number_factors_{}.csv'.format(output_folder, var_name, number_factors))
            df_res = pd.merge(df_res, 
                              df_loadings[['ID'] + ['lambda_factor_{}'.format(ff) for ff in range(1, number_factors + 1)]], 
                              on = 'ID',
                              how = 'left')
            
            # Construct the low-rank component
            df_res['LR_component'] = 0
            for ff in range(1, number_factors + 1):
                df_res['LR_component'] += df_res['factor_{}'.format(ff)] * df_res['lambda_factor_{}'.format(ff)]
                
            # Get results
            res_EV.append({'number_factors': number_factors_EV(df_res, number_factors - 1),
                           'horizon': ii - 1,
                           'cons_adj': cons_adj,
                           'var': var_name})
    res_EV = pd.DataFrame(res_EV)  

    # Get a single table
    latex = pd.pivot_table(res_EV, index = 'cons_adj', columns = 'horizon', values = 'number_factors')
    latex.columns = ['$h = {}$'.format(hh) for hh in range(0, 4 + 1)]
    latex.index = ['Raw Forecasts', 'Adjusted for Time F.E.']
    del latex.columns.name, latex.index.name

    # Export output to CSV for later use
    df_output = latex.T.copy()
    df_output.index = ['SPFfor_Step{}'.format(hh) for hh in range(1, 5 + 1)]
    df_output.to_csv('{}/number_factors_EV_{}.csv'.format(output_folder, output_str))

    # Export to LaTeX
    latex = latex.to_latex(escape = False)
    with open('{}/number_factors_EV_{}.txt'.format(output_folder, output_str), 'w+') as text_file:
        text_file.write(latex)